from enum import IntFlag
from typing import TYPE_CHECKING, Any, TypeVar

from mstrio.api import browsing
from mstrio.connection import Connection
from mstrio.types import ObjectSubTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import Entity, ObjectTypes
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    fetch_objects,
    get_default_args_from_func,
)
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import objects as objects_processors

if TYPE_CHECKING:
    from mstrio.server.project import Project


class ShortcutInfoFlags(IntFlag):
    """Enumeration constants used to specify combination of additional data
    to retrieve with Library Shortcut info."""

    DssDashboardShortcutInfoBookmark = 0b10  # Bookmarks
    DssDashboardShortcutInfoTOC = 0b01  # Table of Contents
    DssDashboardShortcutInfoDefault = 0b00  # None (Default)


class LibraryShortcut(Entity):
    """Representation of a Shortcut referencing an item published to Library
    It points to another object in the metadata and stores additional details
    related to browsing the Library.

    Attributes:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`.
        id (str): ID of the shortcut object
        name (str): Name of the shortcut
        project_id (str): ID of the project that the shortcut is in
        type (ObjectTypes): object type (SHORTCUT_TYPE)
        date_created (DateTime): creation time
        date_modified (DateTime): last modification time
        version (str): object version ID
        acg (Rights): access rights
            (See EnumDSSXMLAccessRightFlags for possible values)
        acl (list[ACE]): object access control list
        hidden (bool): Specifies whether the object is hidden
        owned_by_current_user (bool): whether the current user owns the shortcut
        target (dict): metadata of the target object related to viewing
            in Library
        target_info (dict): general metadata of the target object
        encode_html_content (bool): Flag indicating whether to encode
            HTML content
        current_page_key (str): Current page node key
        stid (int): Shortcut State ID
        current_bookmark (dict): Name and ID of the current bookmark
        prompted (bool): Flag indicating whether the target object has prompts
        datasets_cache_info_hash (str): The hash value for the Dashboard dataset
            cache info
        shortcut_info_flag (bool): flag indicating what information about
            the shortcut should be loaded
    """

    _OBJECT_TYPE = ObjectTypes.SHORTCUT_TYPE
    _OBJECT_SUBTYPES = [ObjectSubTypes.LIBRARY_SHORTCUT]
    _API_GETTERS = {
        (
            'owned_by_current_user',
            'target',
            'encode_html_content',
            'current_page_key',
            'shared_time',
            'last_viewed_time',
            'last_modified_time',
            'stid',
            'current_bookmark',
            'prompted',
            'datasets_cache_info_hash',
            'shortcut_info_flag',
            'dossier_version_hash',
        ): browsing.get_shortcut,
        (
            'name',
            'id',
            'project_id',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
        ): objects_processors.get_info,
    }
    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'shortcut_info_flag': ShortcutInfoFlags}
    _API_PATCH: dict = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'comments',
            'owner',
        ): (
            objects_processors.update,
            'partial_put',
        )
    }

    def __init__(
        self,
        connection: Connection,
        id: str,
        project: 'Project | str | None' = None,
        project_id: str = None,
        project_name: str = None,
        shortcut_info_flag: (
            ShortcutInfoFlags | int
        ) = ShortcutInfoFlags.DssDashboardShortcutInfoTOC,
    ):
        """Initialize the LibraryShortcut object and populate it with
            I-Server data.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`.
            id (str): Shortcut ID
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            shortcut_info_flag (ShortcutInfoFlags, int): flag indicating what
                information about the shortcut should be loaded
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
        )

        super().__init__(
            connection=connection,
            object_id=id,
            project_id=proj_id,
            shortcut_info_flag=get_enum_val(shortcut_info_flag, ShortcutInfoFlags),
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._owned_by_current_user = kwargs.get('owned_by_current_user')
        self._target = kwargs.get('target')
        self._encode_html_content = kwargs.get('encode_html_content')
        self._shortcut_info_flag = (
            ShortcutInfoFlags(kwargs["shortcut_info_flag"])
            if kwargs.get("shortcut_info_flag")
            else None
        )
        self._current_page_key = kwargs.get('current_page_key')
        self._shared_time = kwargs.get('shared_time')
        self._last_viewed_time = kwargs.get('last_viewed_time')
        self._last_modified_time = kwargs.get('last_modified_time')
        self._stid = kwargs.get('stid')
        self._current_bookmark = kwargs.get('current_bookmark')
        self._prompted = kwargs.get('prompted')
        self._datasets_cache_info_hash = kwargs.get('datasets_cache_info_hash')
        self._dashboard_version_hash = kwargs.get('dossier_version_hash')

    T = TypeVar('T')

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Connection,
        to_snake_case: bool = True,
        with_missing_value: bool = False,
    ) -> T:
        """Instantiate a Shortcut from a dict source."""
        return super(Entity, cls).from_dict(
            source=source,
            connection=connection,
            to_snake_case=to_snake_case,
            with_missing_value=with_missing_value,
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ):
        """Alter the properties of a Library Shortcut.

        Args:
            name (str, optional): name of the shortcut object
            description (str, optional): description of the shortcut object
            abbreviation (str, optional): abbreviation of the shortcut object
            hidden (bool, optional): specifies whether the shortcut is hidden
            comments (str, optional): long description of the shortcut
            owner: (str, User, optional): owner ID of the shortcut object
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @property
    def owned_by_current_user(self):
        return self._owned_by_current_user

    @property
    def target(self):
        return self._target

    @property
    def encode_html_content(self):
        return self._encode_html_content

    @property
    def shortcut_info_flag(self):
        return self._shortcut_info_flag

    @property
    def current_page_key(self):
        return self._current_page_key

    @property
    def shared_time(self):
        return self._shared_time

    @property
    def last_viewed_time(self):
        return self._last_viewed_time

    @property
    def last_modified_time(self):
        return self._last_modified_time

    @property
    def stid(self):
        return self._stid

    @property
    def current_bookmark(self):
        return self._current_bookmark

    @property
    def prompted(self):
        return self._prompted

    @property
    def datasets_cache_info_hash(self):
        return self._datasets_cache_info_hash

    @property
    def dashboard_version_hash(self):
        return self._dashboard_version_hash


def get_shortcuts(
    connection: Connection,
    shortcut_ids: list[str],
    project: 'Project | str | None' = None,
    project_id: str = None,
    project_name: str = None,
    shortcut_info_flag: (
        ShortcutInfoFlags | int
    ) = ShortcutInfoFlags.DssDashboardShortcutInfoDefault,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list[dict] | list[LibraryShortcut]:
    """Retrieve information about specific Library shortcuts
    in specific project.

    Args:
        connection (Connection): Connection object
        shortcut_ids (list[str]): ids of target shortcuts
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        shortcut_info_flag (ShortcutInfoFlags | int): a single ShortcutInfoFlags
            that describes what exact info are to be fetched
        to_dictionary (bool): parameter describing output format
        limit (int): limit the number of elements returned.
            If `None` (default), all objects are returned.

    Return:
        list of dictionaries or Shortcut objects,
        depending on `to_dictionary` parameter
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    shortcuts = fetch_objects(
        connection=connection,
        api=browsing.get_shortcuts,
        dict_unpack_value="shortcuts",
        limit=limit,
        filters=filters,
        body=[{"projectId": proj_id, "shortcutIds": shortcut_ids}],
        shortcut_info_flag=get_enum_val(shortcut_info_flag, ShortcutInfoFlags),
    )

    if to_dictionary:
        return shortcuts
    else:
        return [
            LibraryShortcut.from_dict(source=short, connection=connection)
            for short in shortcuts
        ]



from mstrio.api import migration as migrations_api
from mstrio.connection import Connection
from mstrio.utils.helper import rename_dict_keys

REST_ATTRIBUTES_MAP = {
    'packageInfo': 'package_info',
    'importInfo': 'import_info',
}


def get(connection: Connection, id: str) -> dict:
    """Get migration by a specific ID.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the Migration

    Returns:
        dict representing a migration object
    """
    data = migrations_api.get_migration(
        connection, migration_id=id, show_content='default'
    ).json()
    renamed_data = rename_dict_keys(data, REST_ATTRIBUTES_MAP)
    renamed_data['name'] = renamed_data['package_info']['name']
    return renamed_data



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error triggering event {id}")
def trigger_event(connection, id, fields=None, error_msg=None):
    """Trigger an event.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id(str): ID of the event
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    endpoint = f'/api/events/{id}/trigger'
    return connection.post(endpoint=endpoint, params={'fields': fields})


@ErrorHandler(err_msg="Error getting events list.")
def list_events(connection, fields=None, error_msg=None):
    """Get list of events.

    Args:
        connection(object): Strategy One connection object returned by
                `connection.Connection()`.
        error_msg(str, optional): Customized error message.

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(endpoint='/api/events', params={'fields': fields})


@ErrorHandler(err_msg="Error getting event {id} information.")
def get_event(connection, id, fields=None, error_msg=None):
    """Get information of a specific event

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id(str): ID of the event
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.

    Returns:
        HTTP response object returned by the Strategy One REST server
    """
    return connection.get(
        endpoint=f'/api/events/{id}',
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error creating event")
def create_event(connection, body, fields=None, error_msg=None):
    """Create an event.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        body: JSON-formatted body of the new event
        fields:  Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    endpoint = '/api/events'
    return connection.post(endpoint=endpoint, params={'fields': fields}, json=body)


@ErrorHandler(err_msg="Error updating event {id}")
def update_event(connection, id, body, fields=None, error_msg=None):
    """Update an event.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        body: JSON-formatted body of the updated event
        fields:  Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    endpoint = f'/api/events/{id}'
    return connection.put(endpoint=endpoint, params={'fields': fields}, json=body)


@ErrorHandler(err_msg="Error deleting event {id}")
def delete_event(connection, id, error_msg=None):
    """Delete an event.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id: ID of the event to be deleted
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    endpoint = f'/api/events/{id}'
    return connection.delete(endpoint=endpoint)



"""
    This example is for a dashboard user to modify data in a PostgreSQL DB table using
    a modern grid configured with transaction functions. The workflow will rely on this
    script to make actual modifications (delete) in the table.
"""

"""
    The table named USERS has 5 columns USER_ID(Int), USER_LOGIN(Varchar),
    HIRED_DATE(Date), EMAIL(VarChar) and MODIFIED_BY(VarChar), with USER_ID as
    the primary. Data changes for HIRED_DATE(Date), EMAIL(VarChar)
    and MODIFIED_BY(VarChar) are optional.

    In this template we use the library 'psycopg2' (installed in runtime as
    'psycopg2-binary') to connect with the PostgreSQL database, submit atomic requests
    and execute operations. For using other types of database, make sure to use
    their methods correctly.

    Python libraries to handle SQL connections are not installed on the environment by
    default so it is necessary to choose runtime which have such library installed in
    the settings of the script. Additionally it is necessary to specify correct `Network
    Access` to allow script to communicate with some other IP then IP of the
    environment.
"""

import psycopg2
import datetime


try:
    # create connection to database. If you want to use other way then providing
    # user and password please check documentation of this method. When providing
    # passwords or any other sensitive data it is good to use variable of type `secret`.
    connection = None
    connection = psycopg2.connect(
        host=$host,
        port=$port,
        database=$database,
        user=$user_name,
        password=$user_password
    )

    # Prepare parameterized query for select the existing row to delete
    sql_delete_query = """DELETE FROM USERS WHERE USER_ID = %s"""

    # create cursor to execute operations on database
    cursor = connection.cursor()

    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard
    for index in range(len($user_id)):
        # case with DELETE query is the easiest one because we are deleting
        # just based on `USER_ID`
        cursor.execute(sql_delete_query, tuple([int($user_id[index])]))

    # commit all changes when there weren't any errors
    connection.commit()

# error handling
except psycopg2.Error as e:
    # in case of any error we have to rollback all of the changes which
    # has been made on database
    if connection:
        connection.rollback()
    print("Error:", e)

finally:
    # close the database connection
    if connection:
        connection.close()



"""This is the demo script to show how administrator can manage folders, objects
and their dependencies.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.object_management import (
    Folder,
    full_search,
    get_my_personal_objects_contents,
    get_predefined_folder_contents,
    get_search_results,
    get_search_suggestions,
    list_folders,
    list_objects,
    Object,
    PredefinedFolders,
    quick_search,
    quick_search_from_object,
    SearchObject,
    SearchPattern,
    SearchResultsFormat,
    start_full_search,
    quick_search_by_id,
    QuickSearchData
)
from mstrio.object_management.search_enums import SearchDomain, SearchScope
from mstrio.project_objects import Dashboard
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.helpers import Rights

from mstrio.connection import get_connection

# Define variables which can be later used in a script
PROJECT_NAME = $project_name
PROJECT_ID = $project_id

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# list folders from a particular project
list_folders(conn, project_id=PROJECT_ID)

# list configuration-level folders
list_folders(conn)

# get contents of My Personal Objects in a specific folder
get_my_personal_objects_contents(conn, project_id=PROJECT_ID)

# get contents of a pre-defined folder in a specific folder
get_predefined_folder_contents(
    conn, folder_type=PredefinedFolders.PUBLIC_REPORTS, project_id=PROJECT_ID
)

# Define variables which can be later used in a script
FOLDER_NAME = $folder_name
FOLDER_PARENT_ID = $folder_parent_id
FOLDER_DESCRIPTION = $folder_description

# create new folder in a particular folder
new_folder = Folder.create(
    conn, name=FOLDER_NAME, parent=FOLDER_PARENT_ID, description=FOLDER_DESCRIPTION
)

# Define variables which can be later used in a script
FOLDER_ID = $folder_id  # ID for Folder object lookup
TRUSTED_USER_ID = $trusted_user_id

# get folder with a given ID
folder = Folder(conn, id=FOLDER_ID)

# get contents of a folder (optionally as a dictionary)
contents_objs = folder.get_contents()
contents_dict = folder.get_contents(to_dictionary=True)

# alter name and description of a folder
folder.alter(name=FOLDER_NAME, description=FOLDER_DESCRIPTION)

# add ACL right to the folder for the user and propagate to children of folder
folder.acl_add(
    Rights.EXECUTE,
    trustees=TRUSTED_USER_ID,
    denied=False,
    inheritable=True,
    propagate_to_children=True
)  # see utils/acl.py for Rights values

# remove ACL from the folder for the user
folder.acl_remove(Rights.EXECUTE, trustees=TRUSTED_USER_ID, denied=False, inheritable=True)
# see utils/acl.py for Rights values

# Delete folder. When argument `force` is set to `False` (default value), then
# deletion must be approved.
folder.delete(force=True)

# list objects of a given type
reports = list_objects(conn, ObjectTypes.REPORT_DEFINITION)  # see types.py for ObjectTypes values
documents = list_objects(conn, ObjectTypes.DOCUMENT_DEFINITION)

# list objects of a given subtype in the project
cubes = list_objects(conn, ObjectSubTypes.OLAP_CUBE, domain=SearchDomain.PROJECT)

# Define variables which can be later used in a script
OBJECT_ID = $object_id
OBJECT_NAME = $object_name
OBJECT_DESCRIPTION = $object_description
OBJECT_NEW_OWNER = $object_new_owner

# initialize an object of a given type (in this case `FOLDER`) with a given id
object = Object(conn, ObjectTypes.FOLDER, OBJECT_ID)  # see types.py for ObjectTypes values

# alter name and description of an object
object.alter(name=OBJECT_NAME, description=OBJECT_DESCRIPTION)

# alter object's owner field (owner can be passed as a username, user ID, or User object)
object.alter(owner = OBJECT_NEW_OWNER)


# Define variables which can be later used in a script
REPORT_ID = $report_id
DOCUMENT_ID = $document_id
DASHBOARD_ID = $dashboard_id

# certify object
Object(conn, ObjectTypes.REPORT_DEFINITION, REPORT_ID).certify()

# decertify object
Object(conn, ObjectTypes.REPORT_DEFINITION, REPORT_ID).decertify()

# get object properties as a dictionary
object.to_dict()

# create copy of the object in the same folder with the default name
new_object = object.create_copy()

# create copy of the object in the same folder with provided name
new_object = object.create_copy(OBJECT_NAME)

# create copy of the object in another folder with provided name
new_object = object.create_copy(OBJECT_NAME, FOLDER_ID)

# move the object to another folder
object.move(FOLDER_ID)

# Delete objects of a given types (in this case `REPORT` and 'DOCUMENT)
# and given ids. When argument `force` is set to `False` (default value), then
# deletion must be approved.
Object(conn, ObjectTypes.REPORT_DEFINITION, REPORT_ID).delete()
Object(conn, ObjectTypes.DOCUMENT_DEFINITION, DOCUMENT_ID).delete(force=True)

# Define variables which can be later used in a script
SEARCH_OBJECT_ID = $search_object_id
SEARCH_PATTERN = $search_pattern  # string to get search suggestions for

# initialize SearchObject and synchronize with server
search_object = SearchObject(conn, id=SEARCH_OBJECT_ID)

# get search suggestions with the pattern set performed in all unique projects
# across the cluster (it takes effect only in I-Server with cluster nodes)
suggestions = get_search_suggestions(
    conn, project=PROJECT_ID, key=SEARCH_PATTERN, max_results=20, cross_cluster=True
)

# use the stored results of the Quick Search engine to return search results
# and display them as a list (in this particular case all reports which name
# begins with 'A')
objects = quick_search(
    conn,
    PROJECT_ID,
    name='A',
    pattern=SearchPattern.BEGIN_WITH,
    object_types=[ObjectTypes.REPORT_DEFINITION]
)

# perform quick search based on a predefined Search Object (include ancestors
# and acl of returned objects)
objects = quick_search_from_object(
    conn, PROJECT_ID, search_object, include_ancestors=True, include_acl=True
)

# perform quick search based on a project IDs and object IDs
quick_search_data = QuickSearchData(
    project_id=PROJECT_ID,
    object_ids=[OBJECT_ID],
)

objects_by_id = quick_search_by_id(conn, search_data=quick_search_data)

# search the metadata for objects in a specific project that match specific
# search criteria (e.g. super cubes which name ends with `cube`) and save the
# results in I-Server memory
start_dict = start_full_search(
    conn,
    PROJECT_ID,
    name='cube',
    pattern=SearchPattern.END_WITH,
    object_types=ObjectSubTypes.SUPER_CUBE
)

# Retrieve the results of a full metadata search in a tree format which were
# previously obtained by `start_full_search`
results = get_search_results(
    conn, search_id=start_dict['id'], results_format=SearchResultsFormat.TREE
)

# perform full search in one call (instead of the two steps presented above).
# Get all documents which name contains `document` and return in a list format.
# Perform search only in the root folder and its subfolders.
results = full_search(
    conn,
    PROJECT_ID,
    name='document',
    pattern=SearchPattern.CONTAINS,
    object_types=ObjectTypes.DOCUMENT_DEFINITION,
    root=FOLDER_ID
)

# Get object in the project with a given ID using full search
results = full_search(
    conn,
    PROJECT_ID,
    id=OBJECT_ID,
)

# return cubes that are used by the given dashboard (it can be performed with the
# function `full_search` or method `get_connected_cubes` from `Document` class
# or method `get_dependencies` from `Entity` class)
cubes = Dashboard(conn, id=DASHBOARD_ID).get_connected_cubes()
cubes = Dashboard(conn, id=DASHBOARD_ID).list_dependencies(
    project=PROJECT_ID, object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE]
)
cubes = full_search(
    conn,
    project=PROJECT_ID,
    object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE],
    used_by_object_id=DOCUMENT_ID,
    used_by_object_type=ObjectTypes.DOCUMENT_DEFINITION
)

# we can also list cubes that use given dashboard
cubes_using_dashboard = Dashboard(conn, id=DASHBOARD_ID).list_dependents(
    project=PROJECT_ID, object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE]
)

# Define a variable which can be later used in a script
METRIC_ID = $metric_id

# get all objects that use (also recursive) metric with given id
objects = full_search(
    conn,
    project=PROJECT_ID,
    uses_object_id=METRIC_ID,
    uses_object_type=ObjectTypes.METRIC,
    uses_recursive=True
)

# get all managed objects
all_managed_objects = full_search(
    conn,
    project=PROJECT_ID,
    scope=SearchScope.MANAGED_ONLY
)



"""List all the columns for all available tables

1. Connect to the environment using data from workstation
2a. This path is executed when the table is specified in argument of function
3a. Get whole representation of table based on its id
4a. Unpack the response and for the given table get all columns in a specified
    format with  subtype, object id and name
2b. This path is executed when the table is not specified in argument of function
3b. Get all tables in the project specified in the connection object
4b. Get whole representation of each table based on their ids
5b. Unpack the response and for each table return all columns in a specified
    format with subtype, object id and name
"""
from typing import Optional, Union

from mstrio.api import tables
from mstrio.connection import Connection, get_connection
from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.api_helpers import async_get
from mstrio.utils.wip import module_wip, WipLevels

module_wip(globals(), level=WipLevels.PREVIEW)


def list_table_columns(
    connection: Connection, table: Optional[Union[SchemaObjectReference, str]] = None
) -> dict:
    """List all the columns for a given table, if tables is not specified,
    list all the columns for all available tables.

    Args:
        connection: Strategy One connection object returned
            by `connection.Connection()`
        table: SchemaObjectReference of a table or table id

    Returns:
        dictionary of all the columns for a given table, or for all tables
        if table is not specified e.g.
        {'TABLE_NAME':[
            SchemaObjectReference(sub_type='column', object_id='1111',
                                  name='column_name'),
            SchemaObjectReference(sub_type='column', object_id='2222',
                                  name='other_column_name')
        ]}

    """

    def unpack(table_res):
        columns_list_dict = table_res['physicalTable']['columns']
        return [
            SchemaObjectReference(
                col['information']['subType'],
                col['information']['objectId'],
                col['information']['name'],
            )
            for col in columns_list_dict
        ]

    if table:
        table_id = table if isinstance(table, str) else table.object_id
        table_res = tables.get_table(
            connection, table_id, project_id=connection.project_id
        ).json()
        return {table_res['information']['name']: unpack(table_res)}
    else:
        table_res = tables.get_tables(
            connection, project_id=connection.project_id
        ).json()
        ids = [tab['information']['objectId'] for tab in table_res['tables']]
        tables_async = async_get(tables.get_table_async, connection, ids)
        return {table['information']['name']: unpack(table) for table in tables_async}


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# list all the columns for all available tables
list_table_columns(conn)



# Create an administration package with list of active users
# Leave action "use existing"

from mstrio.connection import get_connection


from mstrio.object_management.migration.migration import Migration
from mstrio.object_management.migration.package import (
    PackageConfig,
    PackageContentInfo,
    PackageSettings,
)
from mstrio.users_and_groups.user import list_users

conn = get_connection(workstationData, project_name='Project Analytics')

all_users = list_users(connection=conn)
active_users = [u for u in all_users if u.enabled]

# If no settings are specified, defaults will be used:
# - action: "use existing"
# - acl on replacing objects: "use existing"
# - acl on new objects: "keep acl as source object"
package_settings = PackageSettings()

# For each of the migrated objects, create a PackageContentInfo entry
# An entry can have its own action defined. It will override the default action
# If not specified, the default action will be used
# In this case, we will not set include_dependents to True. This will cause
# I-Server to include all dependent objects, without options to include or
# exclude them.
content_info = [PackageContentInfo(id=obj.id, type=obj.type) for obj in active_users]

package_config = PackageConfig(package_settings, content_info)

my_obj_mig_alt = Migration.create_object_migration(
    connection=conn,
    toc_view=package_config,
    name="test_object_mig",
    project_id=conn.project_id,
)



import logging
import os
from collections.abc import Callable
from datetime import datetime
from hashlib import sha256

from mstrio import config
from mstrio.api import migration as migration_api
from mstrio.connection import Connection
from mstrio.object_management import Object, SearchObject
from mstrio.object_management.migration.package import (
    CATALOG_ITEMS,
    OBJECT_MIGRATION_TYPES_ADMINISTRATION,
    OBJECT_MIGRATION_TYPES_MIN_VERSION,
    OBJECT_MIGRATION_TYPES_OBJECT,
    Action,
    ImportInfo,
    ImportStatus,
    MigratedObjectTypes,
    MigrationPurpose,
    PackageCertificationStatus,
    PackageConfig,
    PackageContentInfo,
    PackageInfo,
    PackageSettings,
    PackageStatus,
    PackageType,
    ProjectMergePackageSettings,
    ProjectMergePackageTocView,
    RequestStatus,
    Validation,
)
from mstrio.server import Environment
from mstrio.server.project import Project
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import DeleteMixin, EntityBase
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import camel_to_snake
from mstrio.utils.progress_bar_mixin import ProgressBarMixin
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import migrations
from mstrio.utils.time_helper import datetime_to_str
from mstrio.utils.version_helper import (
    class_version_handler,
    is_server_min_version,
    method_version_handler,
)

logger = logging.getLogger(__name__)


def list_migration_possible_content(
    connection: Connection, package_type: PackageType
) -> list[ObjectTypes | ObjectSubTypes]:
    """List possible content for migration process based on the package type and
    environment version.

    Args:
        connection (Connection): A Strategy One connection object
        package_type (PackageType): Type of the package

    Returns:
        Set of MigratedObjectTypes"""
    if package_type == PackageType.ADMINISTRATION:
        available_per_package_type = OBJECT_MIGRATION_TYPES_ADMINISTRATION
    elif package_type == PackageType.OBJECT:
        available_per_package_type = OBJECT_MIGRATION_TYPES_OBJECT
    else:
        available_per_package_type = set(MigratedObjectTypes)
    unavailable_per_version = {
        object_type
        for object_type, version in OBJECT_MIGRATION_TYPES_MIN_VERSION.items()
        if not is_server_min_version(connection, version)
    }
    available_migrated_types = available_per_package_type - unavailable_per_version
    types_to_search: list[ObjectTypes | ObjectSubTypes] = []
    for mig_object_type in available_migrated_types:
        obj_type, obj_subtype = CATALOG_ITEMS[mig_object_type]
        if isinstance(obj_subtype, list):
            types_to_search.extend(obj_subtype)
        elif obj_subtype is not None:
            types_to_search.append(obj_subtype)
        else:
            types_to_search.append(obj_type)

    return types_to_search


@method_version_handler(version='11.3.10')
def list_migrations(
    connection: Connection,
    name: str | None = None,
    migration_purpose: MigrationPurpose | str | None = None,
    package_status: PackageStatus | str | None = None,
    import_status: ImportStatus | str | None = None,
    limit: int | None = None,
    to_dictionary: bool = False,
) -> list["Migration"] | list[dict]:
    """Get list of Migration objects.
    Optionally use `to_dictionary` to choose output format.

    Args:
        connection (Connection): A Strategy One connection object
        name (str, optional): characters that the Migration name must contain
        migration_purpose (MigrationPurpose, str, optional): purpose of the
            migration can be either 'object_migration', 'project_merge' or
            'migration_from_shared_file_store'. If None returns migration of all
            purposes
        package_status (PackageStatus, str, optional): status of the migration
            package
        import_status (ImportStatus, str, optional): status of the Migration
            import process
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Metric objects

    Returns:
        List of Migration objects or list of dictionaries
    """
    if migration_purpose:
        objects_ = _list_migrations(
            connection, migration_purpose=migration_purpose, limit=limit
        )
    else:
        all_purposes = ','.join([purpose.value for purpose in MigrationPurpose])
        objects_ = _list_migrations(
            connection,
            migration_purpose=all_purposes,
            limit=limit,
        )

    for obj in objects_:
        obj['name'] = obj['packageInfo']['name']

    if name:
        objects_ = list(filter(lambda x: name in x['name'], objects_))
    if package_status:
        package_status = get_enum_val(package_status, PackageStatus)
        objects_ = list(
            filter(lambda x: package_status == x['packageInfo']['status'], objects_)
        )
    if import_status:
        import_status = get_enum_val(import_status, ImportStatus)
        objects_ = list(
            filter(lambda x: import_status == x['importInfo']['status'], objects_)
        )
    if to_dictionary:
        return objects_
    else:
        return [
            Migration.from_dict(source=obj, connection=connection) for obj in objects_
        ]


def _list_migrations(
    connection: Connection,
    migration_purpose: (
        MigrationPurpose | list[MigrationPurpose] | str | list[str] | None
    ) = None,
    limit: int | None = None,
) -> list[dict]:
    max_call_limit = 1000
    if isinstance(migration_purpose, MigrationPurpose):
        migration_purpose = migration_purpose.value
    if limit is not None and limit <= max_call_limit:
        resp = migration_api.list_migrations(
            connection, limit=limit, migration_type=migration_purpose
        )
        objects_ = resp.json().get('data')
    else:
        resp = migration_api.list_migrations(
            connection, limit=1000, migration_type=migration_purpose
        )
        total_objects = resp.json().get('total')
        if total_objects > max_call_limit:
            objects_ = resp.json().get('data')
            while len(objects_) < total_objects:
                resp = migration_api.list_migrations(
                    connection,
                    limit=1000,
                    offset=len(objects_),
                    migration_type=migration_purpose,
                )
                objects_ += resp.json().get('data')
        else:
            objects_ = resp.json().get('data')
    return objects_


@class_version_handler(version='11.3.10')
class Migration(EntityBase, ProgressBarMixin, DeleteMixin):
    """A class encapsulating migration process from env A to env B.

    Raises VersionException if either of environments are running IServer
    version lower than 11.3.10

    Attributes:
        id(str): ID of the migration
        name(str): Name of the migration
        type(str): MSTR type of the migration object. Returns either
            NOT_SUPPORTED or None
        connection(Connection): A Strategy One connection object
        import_info(ImportInfo): Information about the import process
        package_info(PackageInfo): Information about the package configuration
            and status
        validation(Validation): Information about the validation process
        version(str): Version of the migration API
    """

    _API_GETTERS: dict[str | tuple, Callable] = {
        (
            'id',
            'name',
            'import_info',
            'package_info',
            'validation',
            'version',
        ): migrations.get,
    }
    _API_DELETE = staticmethod(migration_api.delete_migration)

    _FROM_DICT_MAP = {
        'import_info': ImportInfo.from_dict,
        'package_info': PackageInfo.from_dict,
        'validation': Validation.from_dict,
    }

    def __init__(
        self,
        connection: 'Connection',
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        """Initialize migration object by its identifier.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            id (str, optional): identifier of a pre-existing migration.
                Defaults to None.
            name (str, optional): name of a pre-existing migration.
                Defaults to None.
        """
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            migrations = list_migrations(connection=connection, name=name)
            if migrations:
                number_of_objects = len(migrations)
                if number_of_objects > 1:
                    raise ValueError(
                        f"There are {number_of_objects} {type(self).__name__}"
                        " objects with this name. Please initialize with ID."
                    )

                id = migrations[0].to_dict()[  # NOSONAR - shadowing `id` in fn scope
                    'id'
                ]
            else:
                raise ValueError(
                    f"There is no {type(self).__name__} with the given name: '{name}'"
                )
        super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.version = kwargs.get('version')
        self._import_info = kwargs.get('importInfo')
        self._package_info = kwargs.get('packageInfo')
        self._validation = kwargs.get('validation')

    @classmethod
    def create(
        cls,
        connection: 'Connection',
        body: dict,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> 'Migration':
        """Create a totally new migration object.

        Note:
            Project parameters are obsolete when `body` package info is a
            `PackageType.ADMINISTRATION` type.

        Args:
            connection (Connection): A Strategy One connection object
            body (dict): a json body with migration details
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            A new Migration object
        """

        if body['packageInfo']['type'] == PackageType.ADMINISTRATION.value:
            response = migration_api.create_new_migration(
                connection=connection,
                body=body,
                prefer='respond-async',
            )
        else:
            proj_id = get_project_id_from_params_set(
                connection,
                project,
                project_id,
                project_name,
            )
            response = migration_api.create_new_migration(
                connection=connection,
                body=body,
                prefer='respond-async',
                project_id=proj_id,
            )

        response_data = response.json()

        if config.verbose:
            logger.info(
                "Successfully started creation of migration object with ID:"
                f" '{response_data.get('id')}'"
            )

        return cls.from_dict(
            source=camel_to_snake(response_data), connection=connection
        )

    @classmethod
    def create_object_migration(
        cls,
        connection: 'Connection',
        toc_view: dict | PackageConfig,
        tree_view: str | None = None,
        name: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> 'Migration':
        """
        Create a new migration for object migration purpose.

        Args:
            connection (Connection): A Strategy One connection object.
            toc_view (dict, PackageConfig): A dictionary representing the TOC
                view or a PackageConfig object.
            tree_view (str, optional): A string representing the tree view.
            name (str, optional): Name of the migration. Used for identification
                purposes for the convenience of the user. If None default name
                will be generated.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            A new Migration object.
        """
        if isinstance(toc_view, PackageConfig):
            toc_view = toc_view.to_dict()
        tree_view = tree_view or {}
        base_body = Migration._get_body_for_create(
            connection, toc_view, tree_view, name
        )
        base_body['packageInfo']['type'] = PackageType.OBJECT.value
        return cls.create(connection, base_body, project, project_id, project_name)

    @classmethod
    def create_admin_migration(
        cls,
        connection: 'Connection',
        toc_view: dict | PackageConfig,
        tree_view: str | None = None,
        name: str | None = None,
    ) -> 'Migration':
        """
        Create a new migration for administration migration purpose.

        Args:
            connection (Connection): A Strategy One connection object.
            toc_view (dict, PackageConfig): A dictionary representing the TOC
                view or a PackageConfig object.
            tree_view (str): A string representing the tree view.
            name (str, optional): Name of the migration. Used for identification
                purposes for the convenience of the user. If None default name
                will be generated.

        Returns:
            A new Migration object.
        """
        if isinstance(toc_view, PackageConfig):
            toc_view = toc_view.to_dict()
        tree_view = tree_view or {}
        base_body = Migration._get_body_for_create(
            connection, toc_view, tree_view, name
        )
        base_body['packageInfo']['type'] = PackageType.ADMINISTRATION.value
        return cls.create(connection, base_body)

    @classmethod
    def create_project_merge_migration(
        cls,
        connection: 'Connection',
        toc_view: dict | ProjectMergePackageSettings | ProjectMergePackageTocView,
        tree_view: str | None = None,
        name: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> 'Migration':
        """
        Create a new migration for project merge migration purpose.

        Args:
            connection (Connection): A Strategy One connection object.
            toc_view (dict): A dictionary representing the TOC view or a
                ProjectMergePackageSettings | ProjectMergePackageTocView object.
            tree_view (str): A string representing the tree view.
            name (str, optional): Name of the migration. Used for identification
                purposes for the convenience of the user. If None default name
                will be generated.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            A new Migration object.
        """

        if isinstance(toc_view, ProjectMergePackageSettings):
            toc_view = ProjectMergePackageTocView(toc_view).to_dict()
        if isinstance(toc_view, ProjectMergePackageTocView):
            toc_view = toc_view.to_dict()
            toc_view['settings']['aclOnNewObjects'] = [
                toc_view['settings']['aclOnNewObjects']
            ]
        tree_view = tree_view or {}
        base_body = Migration._get_body_for_create(
            connection, toc_view, tree_view, name
        )
        base_body['packageInfo']['type'] = PackageType.OBJECT.value
        base_body['packageInfo']['purpose'] = MigrationPurpose.PROJECT_MERGE.value
        return cls.create(connection, base_body, project, project_id, project_name)

    def reuse(
        self,
        target_env: Connection | Environment,
        target_project: 'Project | str | None' = None,
        target_project_id: str | None = None,
        target_project_name: str | None = None,
    ) -> 'Migration':
        """
        Reuse an already migrated package to create a new one with the same
        properties, and then migrate it to a different environment or
        the same one.

        Args:
            target_env (Connection, Environment): Target environment to migrate
                reused package to.
            target_project (Project | str, optional): Project object or ID or
                name specifying the project. May be used instead of
                `target_project_id` or `target_project_name`.
            target_project_id (str, optional): Project ID
            target_project_name (str, optional): Project name

        Returns:
            A new Migration object based on the reused package.
        """
        target_id = Migration._get_env_id(target_env)
        body = {"importInfo": {"environment": {'id': target_id, 'name': target_id}}}

        if self._package_info.type == PackageType.OBJECT:
            if isinstance(target_env, Environment):
                target_env = target_env.connection
            target_proj_id = get_project_id_from_params_set(
                target_env,
                target_project,
                target_project_id,
                target_project_name,
            )
            target_proj_name = Project(target_env, id=target_proj_id).name
            body['importInfo']['project'] = {
                'id': target_proj_id,
                'name': target_proj_name,
            }

        response_data = migration_api.create_new_migration(
            connection=self.connection, package_id=self._package_info.id, body=body
        ).json()
        mig_dict = migration_api.get_migration(
            self.connection, response_data['id'], show_content='all'
        ).json()
        mig_dict['packageInfo']['replicated'] = True
        migration_api.start_migration(
            target_env,
            prefer='respond-async',
            migration_id=response_data['id'],
            body=mig_dict,
            generate_undo=True,
            project_id=target_project_id,
        ).json()
        if config.verbose:
            logger.info(
                "Successfully reused existing migration object and created new "
                "one with ID:"
                f" '{response_data.get('id')}'"
                " that was already migrated."
            )
        return Migration(connection=self.connection, id=response_data.get('id'))

    @classmethod
    def _upload_binary(
        cls,
        connection: Connection,
        name: str,
        package_type: PackageType | str,
        migration_purpose: MigrationPurpose | str,
        file: bytes,
    ):
        """
        "packageType": str
            ("project", "project_security", "configuration"),
        "packagePurpose": str
            ("object_migration", "project_merge", "migration_group",
            "migration_group_child", "migration_from_shared_file_store")
        """

        package_type = get_enum_val(package_type, PackageType)
        migration_purpose = get_enum_val(migration_purpose, MigrationPurpose)

        conn: Connection = connection
        body = {
            "name": name,
            "extension": "mmp",
            "type": "migrations.packages",
            "size": len(file),
            "sha256": sha256(file).hexdigest(),
            "environment": {
                "id": conn.base_url,
                "name": conn.base_url,
            },
            "extraInfo": {
                "packageType": package_type,
                "packagePurpose": migration_purpose,
            },
        }
        response = migration_api.storage_service_create_file_metadata(connection, body)
        if response.ok:
            cls._file_id = response.json()["id"]
        return migration_api.storage_service_upload_file_binary(
            connection, cls._file_id, file
        )

    @classmethod
    def migrate_from_file(
        cls,
        connection: 'Connection',
        file_path: str,
        package_type: PackageType | str,
        name: str | None = None,
        target_project: 'Project | str | None' = None,
        target_project_id: str | None = None,
        target_project_name: str | None = None,
    ) -> 'Migration':
        """
        Create a new migration object from an existing package file.
        Project information is required in case of object migration.

        Args:
            connection (Connection): A Strategy One connection object.
            file_path (str): A full path to the package file.
            package_type (PackageType | str): Type of the package.
            name (str, optional): Name of the migration. Used for identification
                purposes for the convenience of the user. Defaults to None.
            target_project (Project | str, optional): Project object or ID or
                name specifying the project. May be used instead of
                `target_project_id` or `target_project_name`.
            target_project_id (str, optional): Project ID
            target_project_name (str, optional): Project name

        Returns:
            Migration: A new Migration object.
        """
        target_proj_id = get_project_id_from_params_set(
            connection,
            target_project,
            target_project_id,
            target_project_name,
            assert_id_exists=False,
            no_fallback_from_connection=True,
        )

        package_type = get_enum_val(package_type, PackageType)
        with open(file_path, mode='rb') as file:  # b is important -> binary
            file_content = file.read()

        resp = cls._upload_binary(
            connection=connection,
            name=name,
            package_type=package_type,
            migration_purpose=MigrationPurpose.FROM_FILE,
            file=file_content,
        )
        file_id = resp.json()["id"]
        env_id = Migration._check_slash_in_id(connection.base_url)
        body = {
            "packageInfo": {
                "storage": {
                    "sharedFileStore": {
                        "files": {
                            "id": file_id,
                        }
                    }
                },
                "purpose": MigrationPurpose.FROM_FILE.value,
            },
            "importInfo": {
                "environment": {
                    "id": env_id,
                    "name": env_id,
                }
            },
        }

        if target_proj_id:
            target_proj_name = Project(connection, id=target_proj_id).name

            body['importInfo']['project'] = {
                'id': target_proj_id,
                'name': target_proj_name,
            }
            body['packageInfo']['project'] = {
                'id': target_proj_id,
                'name': target_proj_name,
            }

        response_data = migration_api.create_new_migration(
            connection=connection,
            body=body,
            prefer='respond-async',
            project_id=target_proj_id,
        )
        mig_id = response_data.json().get('id')
        logger.info(f"Successfully started migration from file with ID: '{mig_id}'.")
        return cls.from_dict(
            source=camel_to_snake(response_data.json()), connection=connection
        )

    def get_migration_content(self) -> dict:
        "Get full body of the migration object"
        mig_content = migration_api.get_migration(
            connection=self.connection,
            migration_id=self.id,
            show_content='all',
        ).json()
        return mig_content

    def trigger_validation(
        self,
        target_env: Connection | Environment,
        target_project: 'Project | str | None' = None,
        target_project_id: str | None = None,
        target_project_name: str | None = None,
    ) -> None:
        """Trigger a validate process to migrate the package from the source
        environment to the destination environment according to the action
        defined, without committing any changes to the metadata.
        This API can only be called by administrator when package is created.

        Note:
            Project parameters are obsolete if no project-level objects are
            migrated.

        Args:
            target_env (Connection, Environment): Destination environment
                to validate the package.
            target_project (Project | str, optional): Project object or ID or
                name specifying the project. May be used instead of
                `target_project_id` or `target_project_name`.
            target_project_id (str, optional): Project ID
            target_project_name (str, optional): Project name

        """

        if isinstance(target_env, Environment):
            target_env = target_env.connection

        target_proj_id = get_project_id_from_params_set(
            target_env,
            target_project,
            target_project_id,
            target_project_name,
            assert_id_exists=False,
            no_fallback_from_connection=True,
        )

        target_id = Migration._get_env_id(target_env)
        body = {
            'id': self.id,
            'packageInfo': self._package_info.to_dict(),
            'importInfo': {
                'id': self.import_info.id,
                'environment': {
                    'id': target_id,
                    'name': target_id,
                },
            },
        }

        if target_proj_id:
            body['importInfo']['project'] = {
                'id': target_proj_id,
                'name': Project(target_env, id=target_proj_id).name,
            }

        migration_api.update_migration(
            connection=self.connection, migration_id=self.id, body=body
        )
        fetched_body = migration_api.get_migration(
            connection=self.connection, migration_id=self.id, show_content='all'
        ).json()
        migration_api.trigger_migration_package_validation(
            connection=target_env,
            id=self.id,
            body=fetched_body,
            project_id=target_proj_id,
        )
        status_change = {"validation": {"status": "validating"}}
        migration_api.update_migration(
            connection=self.connection, migration_id=self.id, body=status_change
        )

        if config.verbose:
            logger.info(
                f"Successfully triggered validation process for migration "
                f"with ID: '{self.id}'"
            )

    def certify(
        self,
        status: PackageCertificationStatus | str | None = None,
        creator: User | None = None,
        last_updated_date: datetime | None = None,
        auto_sync: bool | None = None,
    ) -> None:
        """Update a migration's package certification status or trigger a
        process to synchronize the status.

        Notes:
            If argument 'auto_sync' is True, the migration's package
            certification status is synchronized with shared environment via
            storage service.
            Otherwise, the package certification status is updated with the
            definition of 'status', 'creator' and 'last_updated_date'.

        Args:
            status (PackageCertificationStatus, str, optional):
                Package certification status. Optional only if auto_sync is
                set to True.
            creator (User, optional): Creator of this file. Optional only if
                auto_sync is set to True.
            last_updated_date (datetime, optional): The last updated date of the
                certification operation. Optional only if auto_sync is set
                to True.
            auto_sync (bool, optional): If True, the migration's package
                certification status is synchronized with shared environment
                via storage service.
        """

        if all(arg is None for arg in [status, creator, last_updated_date, auto_sync]):
            raise ValueError(
                "Please provide required argument: 'auto_sync' or "
                "arguments: 'status', 'creator' and 'last_updated_date'"
            )

        body = (
            {
                'status': get_enum_val(status, PackageCertificationStatus),
                'operator': {
                    'id': creator.id,
                    'name': creator.name,
                    'fullName': creator.full_name,
                },
                'lastUpdatedDate': datetime_to_str(
                    last_updated_date, '%Y-%m-%dT%H:%M:%S.%f%z'
                ),
            }
            if not auto_sync
            else None
        )
        migration_api.certify_migration_package(
            connection=self.connection,
            id=self.id,
            body=body,
            auto_sync=auto_sync,
        )
        if config.verbose:
            logger.info(
                f"Successfully changed migration status with ID: '{self.id}' "
                f"to status: '{status}'"
            )

    def delete(self, force: bool = False) -> bool:
        """Deletes the Migration.

        Args:
            force: If True, no additional prompt will be shown before deleting
                Migration.

        Returns:
            True for success. False otherwise.
        """
        if self._package_info.purpose == MigrationPurpose.FROM_FILE:
            file_path = self._package_info.storage.path
            storage_files = migration_api.list_storage_files(
                self.connection, file_type='migrations.packages'
            ).json()['data']
            filtered_files = list(
                filter(lambda x: x['path'] == file_path, storage_files)
            )
            if len(filtered_files) > 0:
                migration_api.delete_file_binary(
                    self.connection, filtered_files[0]['id']
                )
        self.package_id = self._package_info.id
        delete_response = super().delete(force=force)
        if not delete_response:
            logger.info(f"Migration with ID: '{self.id}' is already deleted.")
            delete_response = True
        return delete_response

    def download_package(self, save_path: str | None = None) -> dict:
        """Download the package binary to the specified location.

        Args:
            save_path: a full path where the package binary will be saved.
            if None, the package will be saved in the current working directory.

        Returns:
            Dictionary with the filepath and file binary.
        """
        if not save_path:
            save_path = os.getcwd()
        response = migration_api.download_migration_package(
            connection=self.connection, package_id=self.package_info.id
        )
        if response.ok:
            filename = response.headers['Content-Disposition'].split('filename=')[1]
            with open(filename, "wb") as f:
                f.write(response.content)
            filepath = os.path.join(save_path, filename)
            logger.info(f"Package binary saved to: {filepath}")
            return {"filepath": filepath, "file_binary": response.content}

    @classmethod
    def from_dict(
        cls,
        source: dict,
        connection: 'Connection | None' = None,
        to_snake_case: bool = True,
    ):
        """Initialize Migration object from dictionary."""
        data = camel_to_snake(source) if to_snake_case else source.copy()
        return super().from_dict(data, connection, to_snake_case)

    def reverse(
        self,
        target_env: Environment | Connection,
        target_project: 'Project | str | None' = None,
        target_project_id: str | None = None,
        target_project_name: str | None = None,
    ) -> None:
        """Reverse the migration process by importing the undo package.

        Note:
            Project parameters are obsolete if no project-level objects were
            migrated.

        Args:
            target_env (Environment, Connection): Destination environment to
                reverse the migration.
            target_project (Project | str, optional): Project object or ID or
                name specifying the project. May be used instead of
                `target_project_id` or `target_project_name`.
            target_project_id (str, optional): Project ID
            target_project_name (str, optional): Project name
        """
        if isinstance(target_env, Environment):
            target_env = target_env.connection

        target_proj_id = get_project_id_from_params_set(
            target_env,
            target_project,
            target_project_id,
            target_project_name,
            assert_id_exists=False,
            no_fallback_from_connection=True,
        )

        migration_api.update_migration(
            connection=target_env,
            migration_id=self.id,
            body={'importInfo': {'undoRequestStatus': 'requested'}},
            prefer='respond-async',
            project_id=target_proj_id,
        )
        migration_api.update_migration(
            connection=target_env,
            migration_id=self.id,
            body={'importInfo': {'undoRequestStatus': 'approved'}},
            prefer='respond-async',
            project_id=target_proj_id,
        )
        if config.verbose:
            logger.info(f"Successfully reversed migration with ID: '{self.id}'")

    def migrate(
        self,
        target_env: Connection | Environment,
        target_project: 'Project | str | None' = None,
        target_project_id: str | None = None,
        target_project_name: str | None = None,
        generate_undo: bool = True,
    ):
        """Migrate the package to the target environment.
        Project information is required in case of object migration.

        Args:
            target_env (Connection, Environment): Destination environment
                to migrate the package.
            target_project (Project | str, optional): Project object or ID or
                name specifying the project. May be used instead of
                `target_project_id` or `target_project_name`.
            target_project_id (str, optional): Project ID
            target_project_name (str, optional): Project name
            generate_undo (bool, optional): Specify weather to generate an undo
                package or not. True by default.
        """

        if isinstance(target_env, Environment):
            target_env = target_env.connection

        target_proj_id = get_project_id_from_params_set(
            target_env,
            target_project,
            target_project_id,
            target_project_name,
            assert_id_exists=False,
            no_fallback_from_connection=True,
        )

        target_id = Migration._get_env_id(target_env)
        body = {
            "importInfo": {
                "environment": {'id': target_id, 'name': target_id},
                "importRequestStatus": RequestStatus.REQUESTED.value,
            },
        }
        if target_proj_id:
            body['importInfo']['project'] = {
                'id': target_proj_id,
                'name': Project(target_env, id=target_proj_id).name,
            }

        migration_api.update_migration(
            connection=self.connection, migration_id=self.id, body=body
        )
        body['importInfo']['importRequestStatus'] = RequestStatus.APPROVED.value
        migration_api.update_migration(
            connection=self.connection, migration_id=self.id, body=body
        )
        fetched_body = migration_api.get_migration(
            connection=self.connection, migration_id=self.id, show_content='all'
        ).json()
        fetched_body['packageInfo']['replicated'] = True
        response = migration_api.start_migration(
            connection=target_env,
            migration_id=self.id,
            prefer='respond-async',
            generate_undo=generate_undo,
            body=fetched_body,
            project_id=target_proj_id,
        )

        if response.ok and config.verbose:
            logger.info(
                f"Successfully migrated migration '{self.name}' with ID: '{self.id}'"
            )

    def alter_migration_info(
        self,
        name: str | None = None,
        target_env: Connection | Environment | str = None,
        target_project: 'Project | str | None' = None,
        target_project_id: str | None = None,
        target_project_name: str | None = None,
    ) -> None:
        """Alter the migration object info.

        Args:
            name (str, optional): The name of the migration object.
            target_env (Connection | Environment | str, optional): The target
                environment for the migration. It can be a Connection object, an
                Environment object, or a string representing the environment
                    base url.
            target_project (Project | str, optional): The target project for the
                migration. It can be a Project object or a string representing
                the project ID.

        """
        if isinstance(target_env, Environment):
            target_env = target_env.connection

        target_proj_id = get_project_id_from_params_set(
            target_env if isinstance(target_env, Connection) else None,
            target_project,
            target_project_id,
            target_project_name,
            assert_id_exists=False,
            no_fallback_from_connection=True,
        )

        if target_proj_id and isinstance(target_env, Connection):
            target_proj_id = Project(target_env, id=target_proj_id)

        request_body = self._build_body_to_alter_migration_info(
            name=name, target_env=target_env, target_project=target_proj_id
        )
        response = migration_api.update_migration(
            connection=self.connection, migration_id=self.id, body=request_body
        )
        if response.ok and config.verbose:
            logger.info(f"Successfully altered migration object with ID: '{self.id}'")

    @staticmethod
    def _build_body_to_alter_migration_info(
        name: str | None = None,
        target_env: Connection | Environment | str = None,
        target_project: 'Project | str | None' = None,
    ) -> dict:
        request_body = {}
        if name:
            request_body['packageInfo'] = {"name": name}
        if target_env:
            target_id = Migration._get_env_id(target_env)
            request_body['importInfo'] = {
                "environment": {'id': target_id, 'name': target_id}
            }
        if target_project:
            if isinstance(target_project, Project):
                target_project_id = target_project.id
                target_project_name = target_project.name
                request_body['importInfo'] = {
                    "project": {
                        'id': target_project_id,
                        'name': target_project_name,
                    }
                }
            else:
                request_body['importInfo'] = {
                    "project": {
                        'id': target_project,
                    }
                }
        return request_body

    @staticmethod
    def _check_slash_in_id(env_id: str) -> str:
        return env_id if env_id.endswith('/') else f"{env_id}/"

    @staticmethod
    def _get_env_id(env: Connection | Environment | str) -> str:
        if isinstance(env, Connection):
            base_url = env.base_url
        elif isinstance(env, Environment):
            base_url = env.connection.base_url
        else:
            base_url = env

        return Migration._check_slash_in_id(base_url)

    @staticmethod
    def _get_body_for_create(
        connection: 'Connection',
        toc_view: dict,
        tree_view: str,
        name: str | None = None,
    ) -> dict:
        if not name:
            name = Migration._get_new_migration_name()
        env_id = Migration._check_slash_in_id(connection.base_url)
        body = {
            "packageInfo": {
                "name": name,
                "environment": {
                    "id": env_id,
                    "name": env_id,
                },
                "tocView": toc_view,
                "treeView": tree_view,
            },
            "importInfo": {},
        }
        return body

    @staticmethod
    def _get_new_migration_name() -> str:
        return (
            f"New Migration Package {datetime.now().strftime('%m-%d-%Y %I:%M:%S %p')}"
        )

    @staticmethod
    def build_package_config(
        connection: Connection,
        content: list[Object | dict] | SearchObject,
        package_settings: PackageSettings,
        object_action_map: list[tuple] | None = None,
        object_dependents_map: list[tuple] | None = None,
        default_action: Action | None = None,
        default_dependents: bool = False,
    ) -> PackageConfig:
        """
        Build administration or object migration package definition based on the
            provided content.

        Args:
            connection (Connection): A Strategy One connection object.
            content (list[Object | dict] | SearchObject): List of objects to
                migrate or a SearchObject instance.
            package_settings (PackageSettings): Package settings information.
            object_action_map (list[tuple], optional): List of tuples where the
                first element is the object type and the second element is the
                action to perform. If None, default_action will be used.
            object_dependents_map (list[tuple], optional): List of tuples where
                the first element is the object type and the second element is
                the include_dependents flag. If None, default_dependents will be
                used.
            default_action (Action, optional): Default action to apply to the
                list of migration objects passed with the `content` field. When
                an object in the content has the `include_dependents` flag set
                to True, then it's dependents action will be taken from
                `package_settings` field.
                If not provided, the `default_action` from `package_settings`
                will be used.
            default_dependents (bool, optional): Default value for
                include_dependents flag. Defaults to False.

        Returns:
            PackageConfig: A new PackageConfig object.
        """
        default_action = default_action or package_settings.default_action
        content_list = []
        if isinstance(content, SearchObject):
            content = content.run()
        for obj in content:
            if isinstance(obj, dict):
                obj = Object.from_dict(source=obj, connection=connection)
            if object_action_map:
                action = next(
                    (
                        action
                        for obj_type, action in object_action_map
                        if obj.type == obj_type
                    ),
                    default_action,
                )
            else:
                action = default_action
            if object_dependents_map:
                dependents = next(
                    (
                        dependents
                        for obj_type, dependents in object_dependents_map
                        if obj.type == obj_type
                    ),
                    default_dependents,
                )
            else:
                dependents = default_dependents
            content_list.append(
                PackageContentInfo(
                    id=obj.id,
                    type=obj.type,
                    action=action,
                    include_dependents=dependents,
                )
            )
        return PackageConfig(package_settings, content_list)

    @property
    def package_info(self) -> 'PackageInfo':
        return self._package_info

    @property
    def import_info(self) -> 'ImportInfo':
        return self._import_info

    @property
    def validation(self) -> 'Validation':
        self.fetch()
        return self._validation



import logging
from contextlib import suppress
from enum import auto
from typing import TYPE_CHECKING, Any, TypeVar

import pandas as pd

from mstrio import config
from mstrio.connection import Connection
from mstrio.helpers import (
    AggregatedRights,
    IServerError,
    Permissions,
    Rights,
    VersionException,
)
from mstrio.types import ObjectTypes
from mstrio.utils.enum_helper import AutoName, get_enum, get_enum_val
from mstrio.utils.helper import Dictable, exception_handler, filter_obj_list
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import is_server_min_version

if TYPE_CHECKING:
    from mstrio.server import Project
    from mstrio.users_and_groups import UserOrGroup
    from mstrio.utils.entity import Entity

logger = logging.getLogger(__name__)


AGGREGATED_RIGHTS_MAP = {
    Permissions.VIEW: AggregatedRights.VIEW,
    Permissions.MODIFY: AggregatedRights.MODIFY,
    Permissions.FULL_CONTROL: AggregatedRights.ALL,
    Permissions.DENIED_ALL: AggregatedRights.ALL,
    Permissions.DEFAULT_ALL: AggregatedRights.NONE,
    Permissions.CONSUME: AggregatedRights.CONSUME,
}

T = TypeVar("T")


class PropagationBehavior(AutoName):
    """Behavior of ACL propagation to children.
    Allowed values for Folders are:
    - overwrite_children
    - overwrite_recursive
    - append_children
    - append_recursive
    - precise_children
    - precise_recursive
    Allowed values for Users/User Groups are:
    - merge_selected
    - overwrite_selected
    - overwrite_all
    """

    # Folder specific
    OVERWRITE_CHILDREN = auto()
    OVERWRITE_RECURSIVE = auto()
    APPEND_CHILDREN = auto()
    APPEND_RECURSIVE = auto()
    PRECISE_CHILDREN = auto()
    PRECISE_RECURSIVE = auto()
    # User/Group specific
    MERGE_SELECTED = auto()
    OVERWRITE_SELECTED = auto()
    OVERWRITE_ALL = auto()


class ACE(Dictable):
    _FROM_DICT_MAP = {
        'rights': Rights,
    }

    def __init__(
        self,
        deny: bool,
        entry_type: int,
        rights: Rights | int,
        trustee_id: str,
        trustee_name: str,
        trustee_type: int,
        trustee_subtype: int,
        inheritable: bool,
    ):
        """Set ACL object.

        Args:
            deny: Specifies whether access is denied
            entry_type: Access control entry type (1 for object access).
                Possible values can be found in EnumDSSXMLAccessEntryType
            rights: Rights assigned to the designated trustee
            trustee_id: User ID of the designated trustee
            trustee_name: User name of the designated trustee
            trustee_type: Type of the designated trustee
            trustee_subtype: Sub-type of the designated trustee
            inheritable: Specifies whether access control is inherited
        """

        self.rights = rights if isinstance(rights, Rights) else Rights(rights)
        if self.rights not in range(256) and self.rights not in range(
            536_870_912, 536_871_167
        ):
            msg = (
                "Wrong `rights` value, please provide value in range 0-255 or "
                "combination of Rights enums"
            )
            exception_handler(msg)
        self.deny = deny
        self.entry_type = entry_type
        self.trustee_id = trustee_id
        self.trustee_name = trustee_name
        self.trustee_type = trustee_type
        self.trustee_subtype = trustee_subtype
        self.inheritable = inheritable

    def __eq__(self, other: object) -> bool:
        if len(self.__dict__) != len(other.__dict__):
            return False
        return all(
            getattr(self, attr) == getattr(other, attr) for attr in self.__dict__
        )

    @classmethod
    def from_dict(cls, source: dict[str, Any], connection: Connection):
        def translate_names(name: str):
            if name == "type":
                return "entry_type"
            return name

        modified_source = {translate_names(key): val for key, val in source.items()}
        return super().from_dict(modified_source, connection)

    def to_dict(self, camel_case=True):
        def translate_names(name: str):
            if name == "entry_type" or name == "entryType":
                return "type"
            return name

        result_dict = super().to_dict(camel_case=camel_case)
        return {translate_names(key): val for key, val in result_dict.items()}


class ACLMixin:
    """ACLMixin class adds Access Control List (ACL) management for supporting
    objects.

    An ACL is a set of permissions on objects so that users or user groups have
    control over individual objects in the system. Those permissions decide
    whether or not a user can perform a particular class of operations on a
    particular object. For example, a user may have permissions to view and
    execute a report , but cannot modify the report definition or delete the
    report.

    NOTE: Must be mixedin with Entity or its subclasses.
    """

    def list_acl(
        self, to_dataframe: bool = False, to_dictionary: bool = False, **filters
    ) -> pd.DataFrame | list[dict | ACE]:
        """Get Access Control List (ACL) for this object. Optionally filter
        ACLs by specifying filters.

        Args:
            to_dataframe(bool, optional): if True, return datasets as
                pandas DataFrame
            to_dictionary(bool, optional): if True, return datasets as dicts
            **filters: Available filter parameters: [deny, type, rights,
                trustee_id, trustee_name, trustee_type, trustee_subtype,
                inheritable]

        Examples:
            >>> list_acl(deny=True, trustee_name="John")
        """
        acl = filter_obj_list(self.acl, **filters)
        if to_dataframe:
            return pd.DataFrame(acl)
        elif to_dictionary:
            return [acl_obj.to_dict() for acl_obj in acl]
        else:
            return acl

    def acl_add(
        self: "Entity",
        rights: int | Rights | AggregatedRights,
        trustees: "list[UserOrGroup] | UserOrGroup",
        denied: bool = False,
        inheritable: bool | None = None,
        propagate_to_children: bool | None = None,
        propagation_behavior: PropagationBehavior | str | None = None,
    ) -> None:
        """Add Access Control Element (ACE) to the object ACL.

        Note:
            To add rights for the `Root Folder` and the `Freeform Objects`,
            apply the required rights for these folders:
                - Root Folder: D43364C684E34A5F9B2F9AD7108F7828
                - Freeform Objects: 93D8CF3849C1F85DC1A48E8B9E4271F1

            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            rights: The degree to which the user or group is granted or denied
                access to the object. The available permissions are defined in
                `Rights` and `AggregatedRights` Enums
            trustees: list of trustees (`User` or `UserGroup` objects or ids) to
                update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly
            propagation_behavior: Behavior of ACL propagation to children.

        Examples:
            >>> obj.acl_add(rights=Rights.BROWSE | Rights.EXECUTE,
            >>>             trustees=user_obj, denied=True)
        """
        self._update_acl(
            op="ADD",
            rights=rights,
            trustees=trustees,
            denied=denied,
            inheritable=inheritable,
            propagate_to_children=propagate_to_children,
            propagation_behavior=propagation_behavior,
        )

    def acl_remove(
        self: "Entity",
        rights: int | Rights | AggregatedRights,
        trustees: "list[UserOrGroup] | UserOrGroup",
        denied: bool = False,
        inheritable: bool | None = None,
        propagate_to_children: bool | None = None,
        propagation_behavior: PropagationBehavior | str | None = None,
    ) -> None:
        """Remove Access Control Element (ACE) from the object ACL.

        Note:
            To remove rights from the `Root Folder` and the `Freeform Objects`,
            remove them from these folders:
                - Root Folder: D43364C684E34A5F9B2F9AD7108F7828
                - Freeform Objects: 93D8CF3849C1F85DC1A48E8B9E4271F1

            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            rights: The degree to which the user or group is granted or denied
                access to the object. The available permissions are defined in
                `Rights` and `AggregatedRights` Enums
            trustees: list of trustees (`User` or `UserGroup` objects or ids) to
                update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly
            propagation_behavior: Behavior of ACL propagation to children.

        Examples:
            >>> obj.acl_remove(rights=Rights.BROWSE | Rights.EXECUTE,
            >>>                trustees=user_obj, denied=True)
        """
        self._update_acl(
            op="REMOVE",
            rights=rights,
            trustees=trustees,
            denied=denied,
            inheritable=inheritable,
            propagate_to_children=propagate_to_children,
            propagation_behavior=propagation_behavior,
        )

    def acl_alter(
        self: "Entity",
        rights: int | Rights | AggregatedRights,
        trustees: "list[UserOrGroup] | UserOrGroup",
        denied: bool = False,
        inheritable: bool | None = None,
        propagate_to_children: bool | None = None,
        propagation_behavior: PropagationBehavior | str | None = None,
    ) -> None:
        """Alter an existing Access Control Element (ACE) of the object ACL.

        Note:
            To alter rights for the `Root Folder` and the `Freeform Objects`,
            change the required rights for these folders:
                - Root Folder: D43364C684E34A5F9B2F9AD7108F7828
                - Freeform Objects: 93D8CF3849C1F85DC1A48E8B9E4271F1

            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            rights: The degree to which the user or group is granted or denied
                access to the object. The available permissions are defined in
                `Rights` and `AggregatedRights` Enums
            trustees: list of trustees (`User` or `UserGroup` objects or ids) to
                update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly
            propagation_behavior: Behavior of ACL propagation to children.

        Examples:
            >>> obj.acl_alter(rights=Rights.BROWSE | Rights.EXECUTE,
            >>>               trustees=user_obj, denied=True)
        """
        self._update_acl(
            op="REPLACE",
            rights=rights,
            trustees=trustees,
            denied=denied,
            inheritable=inheritable,
            propagate_to_children=propagate_to_children,
            propagation_behavior=propagation_behavior,
        )

    def _update_acl(
        self: "Entity",
        op: str,
        rights: int | Rights | AggregatedRights,
        trustees: "list[UserOrGroup] | UserOrGroup",
        denied: bool = False,
        inheritable: bool | None = None,
        propagate_to_children: bool | None = None,
        propagation_behavior: PropagationBehavior | str | None = None,
    ) -> None:
        """Updates the ACL for this object, performs operation defined by the
        `op` parameter on all objects from `trustees` list.

        Note:
            Argument `propagate_to_children` is used only for objects with
            type `ObjectTypes.FOLDER`.

        Args:
            op: ACL update operator, available values are "ADD", "REMOVE" and
                "REPLACE"
            rights: value of rights to use by the operator
            trustees: list of trustees to update the ACE for
            denied: flag to indicate granted or denied access to the object
            inheritable: Applies only to folders. If set, any objects placed in
                the folder inherit the folder's entry in the ACL.
            propagate_to_children: used for folder objects only, default value
                is None, if set to True/False adds `propagateACLToChildren`
                keyword to the request body and sets its value accordingly
            propagation_behavior: Behavior of ACL propagation to children.
        """

        response = modify_rights(
            connection=self.connection,
            object_type=self._OBJECT_TYPE,
            ids=self.id,
            op=op,
            rights=rights,
            trustees=trustees,
            denied=denied,
            inheritable=inheritable,
            propagate_to_children=propagate_to_children,
            propagation_behavior=propagation_behavior,
        )

        self._set_object_attributes(**response)


class TrusteeACLMixin:
    """TrusteeACLMixin class adds ACL management for Trustee classes.

    Objects currently supporting this Mixin are: (`User` and `UserGroup`).
    """

    def set_permission(
        self: 'UserOrGroup',
        permission: Permissions | str,
        to_objects: str | list[str],
        object_type: 'ObjectTypes | int',
        project: 'Project | str | None' = None,
        propagate_to_children: bool | None = None,
        propagation_behavior: PropagationBehavior | str | None = None,
    ) -> None:
        """Set permission to perform actions on given object(s).

        Function is used to set permission of the trustee to perform given
        actions on the provided objects. Within one execution of the function
        permission will be set in the same manner for each of the provided
        objects. The only available values of permission are: 'View', 'Modify',
        'Full Control', 'Denied All', 'Default All'. Permission is the
        predefined set of rights. All objects to which the rights will be given
        have to be of the same type which is also provided.

        Args:
            permission: The Permission which defines set of rights. See:
                `Permissions` enum
            to_objects: List of object ids on access list for which the
                permissions will be set
            object_type: Type of objects on access list. See: `ObjectTypes` enum
            project: Object or id of Project where the object is
                located. If not passed, Project (project_id) selected in
                Connection object is used
            propagate_to_children: Flag used in the request to determine if
                those rights will be propagated to children of the trustee
            propagation_behavior: Behavior of ACL propagation to children.
        Returns:
            None
        """

        permission = get_enum(permission, Permissions)
        right_value = AGGREGATED_RIGHTS_MAP[permission].value
        denied = permission is Permissions.DENIED_ALL

        # those 2 tries are for clearing current rights (set to default values)
        for is_denied in True, False:
            try:
                current_permission = (
                    Permissions.DENIED_ALL if is_denied else Permissions.FULL_CONTROL
                ).value
                if config.verbose:
                    logger.info(
                        f"Attempting to remove default permission: {current_permission}"
                    )
                modify_rights(
                    connection=self.connection,
                    object_type=object_type,
                    trustees=self.id,
                    op='REMOVE',
                    rights=AggregatedRights.ALL.value,
                    ids=to_objects,
                    denied=is_denied,
                    propagate_to_children=propagate_to_children,
                    propagation_behavior=propagation_behavior,
                    project=project,
                )
                if config.verbose:
                    logger.info(
                        f"Successfully removed permission: {current_permission}"
                    )
            except IServerError:
                pass

        if permission != Permissions.DEFAULT_ALL:
            try:
                if config.verbose:
                    logger.info(f"Attempting to add permission: {permission.value}")
                modify_rights(
                    connection=self.connection,
                    object_type=object_type,
                    trustees=self.id,
                    op='ADD',
                    rights=right_value,
                    ids=to_objects,
                    denied=denied,
                    propagate_to_children=propagate_to_children,
                    propagation_behavior=propagation_behavior,
                    project=project,
                )
                if config.verbose:
                    logger.info(f"Successfully added permission: {permission.value}")
            except IServerError:
                pass

    def set_custom_permissions(
        self: 'UserOrGroup',
        to_objects: str | list[str],
        object_type: 'ObjectTypes | int',
        project: 'Project | str | None' = None,
        execute: str | None = None,
        use: str | None = None,
        control: str | None = None,
        delete: str | None = None,
        write: str | None = None,
        read: str | None = None,
        browse: str | None = None,
    ) -> None:
        """Set custom permissions to perform actions on given object(s).

        Function is used to set rights of the trustee to perform given actions
        on the provided objects. Within one execution of the function rights
        will be set in the same manner for each of the provided objects.
        None of the rights is necessary, but if provided then only possible
        values are 'grant' (to grant right), 'deny' (to deny right), 'default'
        (to reset right) or None which is default value and means that nothing
        will be changed for this right. All objects to which the rights will be
        given have to be of the same type which is also provided.

        Args:
            to_objects: (str, list(str)): List of object ids on access list to
                which the permissions will be set
            object_type (int, ObjectTypes): Type of objects on access list
            project (str, Project): Object or id of Project in which
                the object is located. If not passed, Project
                (project_id) selected in Connection object is used.
            execute (str): value for right "Execute". Available are 'grant',
                'deny', 'default' or None
            use (str): value for right "Use". Available are 'grant',
                'deny', 'default' or None
            control (str): value for right "Control". Available are 'grant',
                'deny', 'default' or None
            delete (str): value for right "Delete". Available are 'grant',
                'deny', 'default' or None
            write  (str): value for right "Write". Available are 'grant',
                'deny', 'default' or None
            read (str): value for right "Read". Available are 'grant',
                'deny', 'default' or None
            browse (str): value for right "Browse. Available are 'grant',
                'deny', 'default' or None
        Returns:
            None
        """

        def modify_custom_rights(
            connection,
            trustee_id: str,
            right: Rights | list[Rights],
            to_objects: list[str],
            object_type: 'ObjectTypes | int',
            denied: bool,
            default: bool = False,
            propagate_to_children: bool | None = None,
            project: 'Project | str | None' = None,
        ) -> None:
            right_value = _get_custom_right_value(right)
            with suppress(IServerError):
                modify_rights(
                    connection=connection,
                    trustees=trustee_id,
                    op='REMOVE',
                    rights=right_value,
                    ids=to_objects,
                    object_type=object_type,
                    project=project,
                    denied=(not denied),
                    propagate_to_children=propagate_to_children,
                )

            op = 'REMOVE' if default else 'ADD'
            with suppress(IServerError):
                modify_rights(
                    connection=connection,
                    trustees=trustee_id,
                    op=op,
                    rights=right_value,
                    ids=to_objects,
                    object_type=object_type,
                    project=project,
                    denied=denied,
                    propagate_to_children=propagate_to_children,
                )

        rights_dict = {
            Rights.EXECUTE: execute,
            Rights.USE: use,
            Rights.CONTROL: control,
            Rights.DELETE: delete,
            Rights.WRITE: write,
            Rights.READ: read,
            Rights.BROWSE: browse,
        }
        if not set(rights_dict.values()).issubset({'grant', 'deny', 'default', None}):
            msg = (
                "Invalid value of the right. Available values are 'grant', 'deny', "
                "'default' or None."
            )
            raise ValueError(msg)

        grant_list = [right for right, value in rights_dict.items() if value == 'grant']
        deny_list = [right for right, value in rights_dict.items() if value == 'deny']
        default_list = [
            right for right, value in rights_dict.items() if value == 'default'
        ]

        modify_custom_rights(
            connection=self.connection,
            trustee_id=self.id,
            right=grant_list,
            to_objects=to_objects,
            object_type=object_type,
            denied=False,
            project=project,
        )
        modify_custom_rights(
            connection=self.connection,
            trustee_id=self.id,
            right=deny_list,
            to_objects=to_objects,
            object_type=object_type,
            denied=True,
            project=project,
        )
        modify_custom_rights(
            connection=self.connection,
            trustee_id=self.id,
            right=default_list,
            to_objects=to_objects,
            object_type=object_type,
            denied=True,
            project=project,
            default=True,
        )


def modify_rights(
    connection: 'Connection',
    object_type: 'ObjectTypes | int',
    ids: list[str] | str,
    op: str,
    rights: int,
    trustees: 'list[UserOrGroup] | UserOrGroup',
    denied: bool = False,
    inheritable: bool | None = None,
    propagate_to_children: bool | None = None,
    propagation_behavior: PropagationBehavior | str | None = None,
    project: 'Project | str | None' = None,
) -> None | dict:
    """Updates the ACL for all given objects specified by id from ids list,
    performs operation defined by the `op` parameter on all objects for
    every user or group from `trustees` list.

    Note:
        Argument `inheritable`, `propagate_to_children`
        and `propagation_behavior` are used only for objects with types:
        - `ObjectTypes.Folder`
        - `ObjectTypes.User`
        - `ObjectTypes.UserGroup`

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        object_type (ObjectTypes, int): Type of every object from ids list.
            One of EnumDSSXMLObjectTypes. Ex. 34 (User or UserGroup),
            44 (Security Role), 32 (Project), 8 (Folder).
        ids (list[str], str): List of object ids or one object id on which
            operations will be performed.
        op (str): ACL update operator, available values are "ADD", "REMOVE"
            and "REPLACE".
        rights (int): Value of rights to use by the operator.
        trustees (list[UserOrGroup], UserOrGroup): List of trustees or one
            trustee to update the ACE for.
        denied (bool): Flag to indicate granted or denied access
            to the object.
        inheritable (bool, optional): Applies only to folders. If set, any
            objects placed in the folder inherit the folder's entry
            in the ACL.
        propagate_to_children (bool, optional): Used for folder objects
            only, default value is None, if set to True/False adds
            `propagateACLToChildren` keyword to the request body and sets
            its value accordingly.
        project (Project, str, optional): Project object or project id
            on which objects are stored, if not specified project id from
            connection will be used.
        propagation_behavior: Behavior of ACL propagation to children.

    Returns:
        Dict with updated object properties if there was only one id in ids
        list or None.
    """
    if not isinstance(ids, list):
        ids = [ids]
    if not isinstance(trustees, list):
        trustees = [trustees]
    if not isinstance(object_type, ObjectTypes):
        object_type = ObjectTypes(object_type)

    trustees = [
        trustee if isinstance(trustee, str) else trustee.id for trustee in trustees
    ]
    project = project.id if project and not isinstance(project, str) else project
    if op not in ["ADD", "REMOVE", "REPLACE"]:
        raise ValueError("Wrong ACL operator passed. Please use ADD, REMOVE or REPLACE")

    for id in ids:
        for trustee in trustees:
            if inheritable is None and object_type is ObjectTypes.FOLDER:
                response = objects_processors.get_info(
                    connection=connection,
                    id=id,
                    object_type=object_type.value,
                    project_id=project,
                )
                tmp = [
                    ace['inheritable']
                    for ace in response.get('acl', [])
                    if ace['trusteeId'] == trustee and ace['deny'] == denied
                ]
                inheritable = False if not tmp else tmp[0]

            body = {
                'acl': [
                    {
                        'op': op,
                        'trustee': trustee,
                        'rights': rights,
                        'type': 1,
                        'denied': denied,
                        'inheritable': inheritable,
                    },
                ]
            }

            if propagate_to_children:
                if propagation_behavior and not is_server_min_version(
                    connection, '11.4.0900'
                ):
                    raise VersionException(
                        "Propagation behavior requires version 11.4.0900 or higher"
                    )

                propagation_behavior = get_enum_val(
                    propagation_behavior, PropagationBehavior
                )

                if is_server_min_version(connection, '11.4.0900') and object_type in [
                    ObjectTypes.USER,
                    ObjectTypes.USERGROUP,
                    ObjectTypes.FOLDER,
                ]:
                    propagation_behavior = propagation_behavior or (
                        'overwrite_recursive'
                        if object_type is ObjectTypes.FOLDER
                        else 'overwrite_all'
                    )
                    body['acl'][0]['propagateToChildren'] = propagate_to_children
                    body['propagateACLToChildren'] = propagate_to_children
                    body['propagationBehavior'] = propagation_behavior
                # On version below 11.4.0900, recursive propagation
                # is only available for ObjectTypes.FOLDER
                elif object_type is ObjectTypes.FOLDER:
                    body['propagateACLToChildren'] = propagate_to_children
                    body['propagationBehavior'] = (
                        propagation_behavior or 'overwrite_recursive'
                    )

            response = objects_processors.update(
                connection=connection,
                id=id,
                body=body,
                object_type=object_type.value,
                project_id=project,
            )
        if len(ids) == 1:
            return response


def _parse_acl_rights_bin_to_dict(rights_bin: int) -> dict[Rights, bool]:
    return {right: rights_bin & right.value != 0 for right in Rights}


def _parse_acl_rights_dict_to_bin(rights_dict: dict[Rights, bool]) -> int:
    output = 0
    for right, given in rights_dict.items():
        if given:
            output |= right.value
    return output


def _get_custom_right_value(right: Rights | list[Rights]) -> int:
    right_value = 0
    if not isinstance(right, list):
        right = [right]
    for r in right:
        if not isinstance(r, Rights):
            try:
                r = Rights[r.upper()]
            except ValueError:
                msg = (
                    f"Invalid custom `right` value: {r}. Available values are: EXECUTE,"
                    f" USE, CONTROL, DELETE, WRITE, READ, BROWSE. See: the Rights enum."
                )
                raise ValueError(msg)
        right_value |= r.value
    return right_value



from typing import TYPE_CHECKING

from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.version_helper import is_server_min_version

if TYPE_CHECKING:
    from mstrio.utils.sessions import FuturesSessionWithRenewal

CUBE_FIELDS = '-data.metricValues.extras,-data.metricValues.formatted'


@ErrorHandler(err_msg="Error getting SQL for report {report_id}.")
def report_sql(
    connection: 'Connection', report_id: str, instance_id: str
) -> 'Response':
    """Get the SQL code for a specific report. This is the SQL that is sent to
    the data warehouse to retrieve the data for the report.

    Args:
        connection: Strategy One REST API connection object
        report_id (str): Unique ID of the report
        instance_id (str): Unique ID of the in-memory instance of a published
            report.

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    return connection.get(
        endpoint=f'/api/v2/reports/{report_id}/instances/{instance_id}/sqlView'
    )


@ErrorHandler(err_msg="Error getting report {report_id} definition. Check report ID.")
def report_definition(connection: 'Connection', report_id: str) -> 'Response':
    """Get the definition of a specific report, including attributes and
    metrics. This in-memory report definition provides information about all
    available objects without actually running any data query/report. The
    results can be used by other requests to help filter large datasets and
    retrieve values dynamically, helping with performance and scalability.

    Args:
        connection: Strategy One REST API connection object
        report_id (str): Unique ID of the report

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    return connection.get(endpoint=f'/api/v2/reports/{report_id}')


@ErrorHandler(err_msg="Error getting report {report_id} contents.")
def report_instance(
    connection: 'Connection',
    report_id: str,
    project_id: str | None = None,
    body: dict | None = None,
    offset: int = 0,
    limit: int = 5000,
    execution_stage: str | None = None,
) -> 'Response':
    """Get the results of a newly created report instance. This in-memory
    report instance can be used by other requests.

    Args:
        connection: Strategy One REST API connection object.
        report_id (str): Unique ID of the report you wish to extract information
            from.
        offset (int, optional): Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.
        execution_stage (str, optional): Execution stage of the report
            Available values: 'resolve_prompts', 'execute_data'

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    if body is None:
        body = {}
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS
    if execution_stage:
        params['executionStage'] = execution_stage
    headers = {'X-MSTR-ProjectID': project_id} if project_id else {}

    return connection.post(
        endpoint=f'/api/v2/reports/{report_id}/instances/',
        json=body,
        params=params,
        headers=headers,
    )


@ErrorHandler(err_msg="Error getting report contents.")
def report_instance_id(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    offset: int = 0,
    limit: int = 5000,
) -> 'Response':
    """Get the results of a previously created report instance, using the in-
    memory report instance created by a POST /api/reports/{reportId}/instances
    request.

    Args:
        connection: Strategy One REST API connection object
        report_id (str): Unique ID of the report you wish to extract information
            from.
        instance_id (str): Unique ID of the in-memory instance of a published
            report.
        offset (int): Optional. Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    return connection.get(
        endpoint=f'/api/v2/reports/{report_id}/instances/{instance_id}', params=params
    )


def report_instance_id_coroutine(
    future_session: "FuturesSessionWithRenewal",
    report_id: str,
    instance_id: str,
    offset: int = 0,
    limit: int = 5000,
):
    """Get the future of a previously created instance for a specific report
    asynchronously, using the in-memory instance created by report_instance().

    Returns:
        Complete Future object.
    """
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(future_session.connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    endpoint = f'/api/v2/reports/{report_id}/instances/{instance_id}'
    future = future_session.get(endpoint=endpoint, params=params)
    return future


@ErrorHandler(err_msg="Error retrieving attribute {attribute_id} elements.")
def report_single_attribute_elements(
    connection: 'Connection',
    report_id: str,
    attribute_id: str,
    offset: int = 0,
    limit: int = 200000,
) -> 'Response':
    """Get elements of a specific attribute of a specific report.

    Args:
        connection: Strategy One REST API connection object.
        report_id (str): Unique ID of the report you wish to extract information
            from.
        attribute_id (str): Unique ID of the attribute in the report.
        offset (int): Optional. Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.
    Returns:
        Complete HTTP response object
    """
    endpoint = f'/api/reports/{report_id}/attributes/{attribute_id}/elements'
    return connection.get(endpoint=endpoint, params={'offset': offset, 'limit': limit})


def report_single_attribute_elements_coroutine(
    future_session: "FuturesSessionWithRenewal",
    report_id: str,
    attribute_id: str,
    offset: int = 0,
    limit: int = 200000,
):
    """Get elements of a specific attribute of a specific report.

    Args:
        future_session(object): Future Session object to call Strategy One REST
            Server asynchronously
        report_id (str): Unique ID of the report you wish to extract information
            from.
        attribute_id (str): Unique ID of the attribute in the report.
        offset (int): Optional. Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As
            an example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.
    Returns:
        Complete Future object
    """
    endpoint = f'/api/reports/{report_id}/attributes/{attribute_id}/elements'
    future = future_session.get(
        endpoint=endpoint, params={'offset': offset, 'limit': limit}
    )
    return future


@ErrorHandler(err_msg="Error getting collection of prompts for report {report_id}")
def get_report_prompts(
    connection: 'Connection',
    report_id: str,
    closed: bool | None = None,
    fields: str | None = None,
) -> 'Response':
    """Get the collection of prompts and their respective definitions from a
    report.

    Args:
        connection: Strategy One REST API connection object.
        report_id (str): Unique ID of the report you wish
            to extract information from.
        closed: Prompt status, true means get closed prompt,
            false means get open prompt
        fields: Comma-separated, top-level field whitelist
            that allows the client to selectively retrieve
            part of the response model.

    """
    endpoint = f'/api/reports/{report_id}/prompts'
    return connection.get(
        endpoint=endpoint, params={'closed': closed, 'fields': fields}
    )


@ErrorHandler(err_msg="Error providing prompt answers for report {report_id}.")
def answer_report_prompts(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    body: dict,
    project_id: str | None = None,
) -> 'Response':
    """Provide answers to the prompts in a report instance.

    Args:
        connection (Connection): Strategy One REST API connection object.
        report_id (str): Unique ID of the report
        instance_id (str): Unique ID of the in-memory instance of a published
            report
        body (dict): dict containing the prompt answers
            A proper body for this request needs the following structure
            (answers are only needed if useDefault is False)
            (only one between key, name and id is required):
                {
                    "prompts": [
                        {
                            "key": "prompt_key",
                            "name": "prompt_name",
                            "id": "prompt_id",
                            "type": "prompt_type",
                            "useDefault": False,
                            "answers": [],
                        }
                    ]
                }
            The available prompt types are:
                    UNSUPPORTED, VALUE, ELEMENTS, EXPRESSION, OBJECTS, LEVEL
        project_id (str, optional): Project ID.

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    if not project_id:
        project_id = connection.project_id
    endpoint = f'/api/reports/{report_id}/instances/{instance_id}/prompts/answers'
    return connection.put(
        endpoint=endpoint, json=body, headers={'X-MSTR-ProjectID': project_id}
    )


@ErrorHandler(err_msg="Error getting prompted report {report_id} instance.")
def get_prompted_instance(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    closed: bool | None = None,
    fields: str | None = None,
    project_id: str | None = None,
) -> 'Response':
    """Get the collection of prompts and their respective definitions from a
    report instance. This endpoint will return data only when the report
    instance has prompt which need to be answered.

    Args:
        connection: Strategy One REST API connection object.
        report_id (str): Unique ID of the report you wish
            to extract information from.
        instance_id (str): Unique ID of the in-memory instance of a published
            report.
        closed(bool, optional): Prompt status, true means get closed prompt,
            false means get open prompt
        fields (str, optional): Comma-separated, top-level field whitelist
            that allows the client to selectively retrieve
            part of the response model.

    """
    if not project_id:
        project_id = connection.project_id
    endpoint = f'/api/reports/{report_id}/instances/{instance_id}/prompts'
    return connection.get(
        endpoint=endpoint,
        params={'closed': closed, 'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(
    err_msg="Error getting prompted report {report_id} attribute element prompt."
)
def get_report_attribute_element_prompt(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    prompt_id: str,
    offset: int = 0,
    limit: int = 100,
    search_pattern: str | None = None,
    fields: str | None = None,
) -> 'Response':
    """Get available attribute element for attribute element prompt"""
    endpoint = (
        f'/api/reports/{report_id}/instances/{instance_id}/prompts/{prompt_id}/elements'
    )
    return connection.get(
        endpoint=endpoint,
        params={
            'offset': offset,
            'limit': limit,
            'searchPattern': search_pattern,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error getting status of report instance {instance_id}")
def get_report_status(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    project_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> 'Response':
    """Get the status of a report instance.

    Args:
        connection (Connection): Strategy One REST API connection object.
        report_id (str): Report ID
        instance_id (str): Report Instance ID
        project_id (str, optional): Project ID
        fields (str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/reports/{report_id}/instances/{instance_id}/status'
    headers = {'X-MSTR-ProjectID': project_id} if project_id else {}

    return connection.get(
        endpoint=endpoint,
        headers=headers,
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error getting VLDB settings for report with ID: {id}")
def get_vldb_settings(
    connection: 'Connection',
    id: str,
    instance_id: str | None = None,
    error_msg: str | None = None,
):
    """Get advanced VLDB settings for a report.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Datasource ID
        instance_id (string, optional): Report Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/model/reports/{id}?showAdvancedProperties=true',
        headers={'X-MSTR-MS-Instance': instance_id},
    )


@ErrorHandler(err_msg="Error updating VLDB settings for report with ID {id}")
def update_vldb_settings(
    connection: 'Connection',
    id: str,
    body: dict,
    instance_id: str,
    project_id: str | None = None,
    error_msg: str | None = None,
):
    """Update metadata of advanced VLDB settings for a report.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Report ID
        instance_id (string): Report Instance ID
        body (dict): JSON-formatted data used to update VLDB settings
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.put(
        endpoint=f'/api/model/reports/{id}',
        headers={'X-MSTR-MS-Instance': instance_id, 'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error getting metadata of VLDB settings for report with ID {id}")
def get_applicable_vldb_settings(
    connection: 'Connection',
    id: str,
    instance_id: str | None = None,
    error_msg: str | None = None,
):
    """Get metadata of advanced VLDB settings for a report.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Report ID
        instance_id (string, optional): Report Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/model/reports/{id}/applicableVldbProperties',
        headers={'X-MSTR-MS-Instance': instance_id},
    )


@ErrorHandler(err_msg="Error executing report {report_id}.")
def execute_report(
    connection: 'Connection',
    report_id: str,
    prefer: str = 'respond-async',
    instance_id: str | None = None,
    project_id: str | None = None,
    execution_stage: str | None = None,
    error_msg: str | None = None,
):
    """Execute a report. A project_id is required to execute the request if
    instance_id is not provided in the header. Set "Prefer" to respond-async to
    execute this API asynchronously.

    Args:
        connection (Connection): Strategy One REST API connection object.
        report_id (str): Unique ID of the report you wish to execute.
        prefer (str, optional): Response preference. Default is 'respond-async'.
        instance_id (str, optional): Unique ID of an existing in-memory instance
            of a published report.
        project_id (str, optional): Unique ID of the project containing the
            report.
        execution_stage (str, optional): Execution stage the report is executed
            to. Available values: 'resolve_prompts', 'execute_data', 'no_action'
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    headers = {'Prefer': prefer}
    params = {}
    if project_id:
        headers['X-MSTR-ProjectID'] = project_id
    if instance_id:
        headers['X-MSTR-MS-Instance'] = instance_id
    if execution_stage:
        params['executionStage'] = execution_stage
    return connection.post(
        endpoint=f'/api/model/reports/{report_id}/instances',
        headers=headers,
        params=params,
    )



from mstrio.api import datasources as datasources_api
from mstrio.connection import Connection
from mstrio.utils import helper


def get_mappings(
    connection: Connection,
    project_id: str | None = None,
    limit: int | None = None,
    default_connection_map: bool = False,
    filters: str = None,
):
    helper.validate_param_value('limit', limit, int, min_val=1, special_values=[None])
    response = datasources_api.get_datasource_mappings(
        connection=connection,
        default_connection_map=default_connection_map,
        project_id=project_id,
    )
    if response.ok:
        response = response.json()
        for mapping in response['mappings']:
            if 'locale' not in mapping:
                mapping['locale'] = {'name': '', 'id': ''}
        mappings = helper._prepare_objects(response, filters, 'mappings', project_id)
        if limit:
            mappings = mappings[:limit]
        return mappings
    else:
        return []


def update_project_datasources(connection: Connection, id: str, body: dict):
    return (
        datasources_api.update_project_datasources(
            connection=connection, id=id, body=body
        )
        .json()
        .get('datasources')
    )


def execute_query(connection: Connection, body: dict, id: str, project_id: str):
    return datasources_api.execute_query(
        connection=connection, id=id, project_id=project_id, body=body
    ).json()


def get_query_results(connection: Connection, id: str):
    return datasources_api.get_query_results(connection=connection, id=id).json()



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting privileges.")
def get_privileges(connection, error_msg=None):
    """Get the set of available privileges for the platform.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/privileges',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting privilege categories.")
def get_privilege_categories(connection, error_msg=None):
    """Get the set of available privilege categories for the platform.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/privileges/categories',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting information for set of security roles.")
def get_security_roles(connection, fields=None, error_msg=None):
    """Get information for all security roles. A security role describes the
    ability to do something, such as create, edit, add, delete, view, manage,
    save, search, share, export, and so on. A security role has a name, a
    description, and a privilege.

    Args:
        connection: Strategy One REST API connection object
        fields: top-level field whitelist which allows client to selectively
            retrieve part of the response model
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/securityRoles',
        headers={'X-MSTR-ProjectID': None},
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error creating new security role.")
def create_security_role(connection, body, error_msg=None):
    """Create a new security role.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/securityRoles',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting security role {id} information.")
def get_security_role(connection, id, error_msg=None):
    """Get information for a security role with security role Id.

    Args:
        connection: Strategy One REST API connection object
        id (string): Security role ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/securityRoles/{id}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error deleting security role with ID {id}.")
def delete_security_role(connection, id, error_msg=None):
    """Delete info for a security role with given Id.

    Args:
        connection: Strategy One REST API connection object
        id (string): Security role ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(
        endpoint=f'/api/securityRoles/{id}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating security role with ID {id}")
def update_security_role(connection, id, body, error_msg=None):
    """Update information for a specific security role.

    Args:
        connection: Strategy One REST API connection object
        id (string): Security role ID
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint=f'/api/securityRoles/{id}',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(
    err_msg="Error getting security role with ID {id} for project with ID {project_id}"
)
def get_security_role_for_project(connection, id, project_id, error_msg=None):
    """Get all users and user groups that are linked to a specific security
    role.

    Args:
        connection: Strategy One REST API connection object
        id (string): Security role ID
        project_id (string, optional): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/securityRoles/{id}/projects/{project_id}/members',
        headers={'X-MSTR-ProjectID': None},
    )



"""
Get subscriptions by user emails.

1. Connect to the environment using data from workstation
2. Specify emails to search for in subscriptions
3. Get all subscriptions and iterate over them
4. Iteration over all user's recipients of the subscription
5. Try to get the addresses of the user from the response with subscriptions
6. If the user is not in the response with subscriptions get its addresses
7. Iterate over all email addresses of the user and print message with
   information about the subscription name and id and user email
"""

from mstrio.connection import Connection, get_connection
from mstrio.distribution_services import list_subscriptions
from mstrio.utils.response_processors import users


def get_subscriptions_by_emails(connection: 'Connection', emails: list[str]) -> None:
    """Get subscriptions by user emails.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        emails (list[str]): List of emails to search for in subscriptions.
    """
    user_addresses = {}
    for sub in list_subscriptions(connection=connection):
        # Iterate over all user recipients of the subscription
        for user_id in [x['id'] for x in sub.recipients if x['type'] == 'user']:
            try:
                # Try to get the addresses of the user from the dictionary
                # to reduce the number of API calls
                addresses = user_addresses[user_id]
            except KeyError:
                # If the user is not in the dictionary, get the addresses from the API
                addresses = users.get_addresses(conn, id=user_id)['addresses']
                # Store the addresses in the dictionary for future use
                user_addresses[user_id] = addresses
            # Iterate over all email addresses of the user
            for email in [
                addr['physicalAddress']
                for addr in addresses
                if addr['deliveryType'] == 'email' and addr['physicalAddress'] in emails
            ]:
                print(
                    f'Found in subscription: {sub.name} with ID: {sub.id}, '
                    f'user email: {email}'
                )


# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

emails_to_find = ['user_1@email.com', 'user_2@email.com']

get_subscriptions_by_emails(conn, emails_to_find)



"""This is the demo script to show how administrator can manage server and the
cluster with its nodes.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.server import Cluster, Environment, Project

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# get the cluster for given connection
clstr = Cluster(connection=conn)

# save topologies of services or nodes from this cluster into a dataframe
nodes_topology_df = clstr.nodes_topology()
services_topology_df = clstr.services_topology()

# get list of services grouped by nodes or by services
services_by_nodes = clstr.list_services(group_by='nodes')
services_by_services = clstr.list_services(group_by='services')

# Define variables which can be used in a script
NODE_NAME = $node_name
NODE_NAME_2 = $node_name_2

# get a list of nodes (information about projects within each node is given there)
nodes = clstr.list_nodes()
# get a Node from the list and get its properties
node = nodes[0]

node.name
node.address
node.service_control
node.port
node.status
node.load
node.projects
node.default

# get a list of nodes as a dictionary
nodes = clstr.list_nodes(to_dictionary=True)
# remove/add a node from/to a cluster (node with the given name should exist)
clstr.remove_node(node=NODE_NAME)
clstr.add_node(node=NODE_NAME)

# remove/add a node from/to a cluster and remove/add it to a cluster startup membership
clstr.remove_node(node=NODE_NAME, add_to_cluster_startup=True)
clstr.add_node(node=NODE_NAME, remove_from_cluster_startup=True)

# get a list of cluster startup membership
clstr.list_cluster_startup_membership()

# update a list of cluster startup membership
clstr.update_cluster_startup_membership(nodes=[NODE_NAME, NODE_NAME_2])

# get name of default (primary) node of the cluster and set new default node
clstr.default_node
clstr.set_primary_node(node=NODE_NAME)

# update node settings or reset them to default values
clstr.update_node_settings(
    node=NODE_NAME, load_balance_factor=99, initial_pool_size=511, max_pool_size=1023
)
clstr.reset_node_settings(node=NODE_NAME)

# Define a variable which can be later used in a script
SERVICE_NAME = $service_name

# stop/start service on selected nodes without providing login and password,
# so user will be prompted to provide them (error will be thrown in case of wrong
# names of service or nodes)
clstr.stop(service=SERVICE_NAME, nodes=[NODE_NAME, NODE_NAME_2])
clstr.start(service=SERVICE_NAME, nodes=[NODE_NAME, NODE_NAME_2])

# Alternatively, the user can provide login and password as arguments to the methods
SSH_LOGIN = $ssh_login
SSH_PASSWORD = $ssh_password

clstr.stop(service=SERVICE_NAME, nodes=[NODE_NAME, NODE_NAME_2], login=SSH_LOGIN,
           passwd=SSH_PASSWORD)
clstr.start(service=SERVICE_NAME, nodes=[NODE_NAME, NODE_NAME_2], login=SSH_LOGIN,
            passwd=SSH_PASSWORD)

# list all projects available for the given connection (it is possible via
# class Cluster or Environment)
env = Environment(connection=conn)
projects = env.list_projects()
projects = clstr.list_projects()

# load or unload chosen project (it is possible via class Cluster or
# Project)
project = Project(connection=conn, name=PROJECT_NAME)
project.load()
project.unload()

# get status of the project on the specified node
project.get_status_on_node(node=NODE_NAME)

# via Cluster can we also specify on which node(s) project will be loaded
# or unloaded
clstr.load_project(project=PROJECT_NAME, on_nodes=[NODE_NAME, NODE_NAME_2])
clstr.unload_project(project=PROJECT_NAME, on_nodes=[NODE_NAME, NODE_NAME_2])

# get settings of a server as a dataframe
server_settings_df = env.server_settings.to_dataframe

# Define a variable which can be later used in a script
FILE_NAME = $file_name  # file name with extension 'csv', 'json' or 'pickle'

# save/load settings of a server to/from a file (format can be 'csv', 'json' or
# 'pickle')
env.server_settings.to_csv(name=FILE_NAME)
env.server_settings.import_from(file=FILE_NAME)

# update some settings of a server
env.server_settings.allowUserLoginWithFullName = True
env.server_settings.minCharsPasswordChanges = 1
env.server_settings.enableUserManagementWithoutEmail = True
env.server_settings.update()



from typing import Any, TypeVar

from mstrio.connection import Connection
from mstrio.object_management.folder import Folder

# Moved to library_shortcuts (imported for backwards compatibility)
from mstrio.object_management.library_shortcut import (  # noqa: F401
    ShortcutInfoFlags,
    get_shortcuts,
)
from mstrio.object_management.search_enums import SearchDomain
from mstrio.object_management.search_operations import full_search
from mstrio.server.project import Project
from mstrio.types import ObjectSubTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin, ObjectTypes
from mstrio.utils.helper import get_default_args_from_func
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import objects as objects_processors


def list_shortcuts(
    connection: Connection,
    name: str | None = None,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    folder: Folder | None = None,
    folder_id: str | None = None,
    folder_path: str | None = None,
    **filters,
):
    """List all shortcuts in a project.

    The project may be specified by either `project_id`, `project_name` or
        `project`. If the project is not specified in either way, the project
        from the `connection` object is used.

    Args:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`.
        name (string, optional): The search pattern for listing shortcuts.
            Supports wildcards '*' (any number of characters) and '?' (exactly
            one character).
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True, the method will return
            dictionaries with the shortcuts' properties instead of Shortcut
            objects. Defaults to False.
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        folder (Folder, optional): Folder object specifying where the search
            will be performed.
        folder_id (string, optional): ID of a folder where the search
            will be performed. Can be provided as an alternative to `folder`
            parameter.
        folder_path (str, optional): Path of the folder in which the search
            will be performed. Can be provided as an alternative to `folder`
            and `folder_id` parameters. The path has to be provided in the
            forward-slash format, beginning with project name, e.g.
            "/MicroStrategy Tutorial/Public Objects/Metrics".
        **filters: Available filter parameters: ['id', 'type', 'subtype',
            'date_created', 'date_modified', 'version', 'owner', 'ext_type',
            'view_media', 'certified_info']

    """

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    validate_owner_key_in_filters(filters)

    if folder and not folder_id:
        folder_id = folder.id
    if folder_path and folder_id:
        folder_path = None

    # No endpoint for listing shortcuts. Using full_search instead.
    objects = full_search(
        connection,
        object_types=ObjectSubTypes.SHORTCUT,
        domain=SearchDomain.PROJECT,
        project=proj_id,
        name=name,
        limit=limit,
        root=folder_id,
        root_path=folder_path,
        **filters,
    )

    if to_dictionary:
        return objects

    return Shortcut.bulk_from_dict(source_list=objects, connection=connection)


class Shortcut(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """Representation of a Shortcut object. It points to another object in
    the metadata.
    DEPRECATION: This class no longer provides information about details of the
        target object's state in Library. For those details, please use
        `LibraryShortcut`.

    Attributes:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`.
        id (str): ID of the shortcut object
        name (str): Name of the shortcut
        project_id (str): ID of the project that the shortcut is in
        type (ObjectTypes): object type (SHORTCUT_TYPE)
        date_created (DateTime): creation time
        date_modified (DateTime): last modification time
        version (str): object version ID
        acg (Rights): access rights
            (See EnumDSSXMLAccessRightFlags for possible values)
        acl (list[ACE]): object access control list
        hidden (bool): Specifies whether the object is hidden
        target_info (dict): general metadata of the target object
    """

    _OBJECT_TYPE = ObjectTypes.SHORTCUT_TYPE
    _OBJECT_SUBTYPES = [ObjectSubTypes.SHORTCUT]
    _API_GETTERS = {
        (
            'name',
            'id',
            'project_id',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
        ): objects_processors.get_info,
    }
    _API_PATCH: dict = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'comments',
            'owner',
        ): (
            objects_processors.update,
            'partial_put',
        )
    }

    def __init__(
        self,
        connection: Connection,
        id: str,
        project: 'Project | str | None' = None,
        project_id: str = None,
        project_name: str = None,
    ):
        """Initialize the Shortcut object and populate it with I-Server data.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`.
            id: Shortcut ID
            project_id: ID of the project that the shortcut is in
        """

        if id is None:
            raise AttributeError("Please specify 'id' parameter in the constructor.")
        else:
            proj_id = get_project_id_from_params_set(
                connection,
                project,
                project_id,
                project_name,
            )

            super().__init__(
                connection=connection,
                object_id=id,
                project_id=proj_id,
            )

    T = TypeVar('T')

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Connection,
        to_snake_case: bool = True,
        with_missing_value: bool = False,
    ) -> T:
        """Instantiate a Shortcut from a dict source."""
        return super(Entity, cls).from_dict(
            source=source,
            connection=connection,
            to_snake_case=to_snake_case,
            with_missing_value=with_missing_value,
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ):
        """Alter the shortcut's properties.

        Args:
            name (str, optional): name of the shortcut object
            description (str, optional): description of the shortcut object
            abbreviation (str, optional): abbreviation of the shortcut object
            hidden (bool, optional): specifies whether the shortcut is hidden
            comments (str, optional): long description of the shortcut
            owner: (str, User, optional): owner of the shortcut object
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def create_shortcut(
        self,
        target_folder_id=None,
        target_folder_path=None,
        target_folder=None,
        project=None,
        project_id=None,
        project_name=None,
        to_dictionary=False,
    ):
        raise ValueError("Shortcut cannot refer to another shortcut.")
